# -*- coding: utf-8 -*-

"""Test __len__ for .NET classes implementing ICollection/ICollection<T>."""

import System
import pytest
from Python.Test import MpLengthCollectionTest, MpLengthExplicitCollectionTest, MpLengthGenericCollectionTest, MpLengthExplicitGenericCollectionTest

def test_simple___len__():
    """Test __len__ for simple ICollection implementers"""
    import System
    import System.Collections.Generic
    l = System.Collections.Generic.List[int]()
    assert len(l) == 0
    l.Add(5)
    l.Add(6)
    assert len(l) == 2

    d = System.Collections.Generic.Dictionary[int, int]()
    assert len(d) == 0
    d.Add(4, 5)
    assert len(d) == 1

    a = System.Array[int]([0,1,2,3])
    assert len(a) == 4

def test_custom_collection___len__():
    """Test __len__ for custom collection class"""
    s = MpLengthCollectionTest()
    assert len(s) == 3

def test_custom_collection_explicit___len__():
    """Test __len__ for custom collection class that explicitly implements ICollection"""
    s = MpLengthExplicitCollectionTest()
    assert len(s) == 2

def test_custom_generic_collection___len__():
    """Test __len__ for custom generic collection class"""
    s = MpLengthGenericCollectionTest[int]()
    s.Add(1)
    s.Add(2)
    assert len(s) == 2

def test_custom_generic_collection_explicit___len__():
    """Test __len__ for custom generic collection that explicity implements ICollection<T>"""
    s = MpLengthExplicitGenericCollectionTest[int]()
    s.Add(1)
    s.Add(10)
    assert len(s) == 2

def test_len_through_interface_generic():
    """Test __len__ for ICollection<T>"""
    import System.Collections.Generic
    l = System.Collections.Generic.List[int]()
    coll = System.Collections.Generic.ICollection[int](l)
    assert len(coll) == 0

def test_len_through_interface():
    """Test __len__ for ICollection"""
    import System.Collections
    l = System.Collections.ArrayList()
    coll = System.Collections.ICollection(l)
    assert len(coll) == 0



# -*- coding: utf-8 -*-

"""Test CLR class constructor support."""

import System


def test_enum_constructor():
    """Test enum constructor args"""
    from System import TypeCode
    from Python.Test import EnumConstructorTest

    ob = EnumConstructorTest(TypeCode.Int32)
    assert ob.value == TypeCode.Int32


def test_flags_constructor():
    """Test flags constructor args"""
    from Python.Test import FlagsConstructorTest
    from System.IO import FileAccess

    flags = FileAccess.Read | FileAccess.Write
    ob = FlagsConstructorTest(flags)
    assert ob.value == flags


def test_struct_constructor():
    """Test struct constructor args"""
    from System import Guid
    from Python.Test import StructConstructorTest

    guid = Guid.NewGuid()
    ob = StructConstructorTest(guid)
    assert ob.value == guid


def test_subclass_constructor():
    """Test subclass constructor args"""
    from Python.Test import SubclassConstructorTest

    class Sub(System.Exception):
        pass

    instance = Sub()
    ob = SubclassConstructorTest(instance)
    assert isinstance(ob.value, System.Exception)


def test_multiple_constructor():
    from Python.Test import MultipleConstructorsTest
    import System

    # Test parameterless
    ob = MultipleConstructorsTest()
    assert ob.value == ""



# -*- coding: utf-8 -*-
# FIXME: This test module randomly passes/fails even if all tests are skipped.
# Something fishy is going on with the Test fixtures. Behavior seen on CI on
# both Linux and Windows
# TODO: Remove delay of class creations. Adding SetUp/TearDown may help

"""Test sub-classing managed types"""

import System
import pytest
from Python.Test import (IInterfaceTest, SubClassTest, EventArgsTest,
                         FunctionsTest)
from System.Collections.Generic import List


def interface_test_class_fixture(subnamespace):
    """Delay creation of class until test starts."""

    class InterfaceTestClass(IInterfaceTest):
        """class that implements the test interface"""
        __namespace__ = "Python.Test." + subnamespace

        def foo(self):
            return "InterfaceTestClass"

        def bar(self, x, i):
            return "/".join([x] * i)

    return InterfaceTestClass


def derived_class_fixture(subnamespace):
    """Delay creation of class until test starts."""

    class DerivedClass(SubClassTest):
        """class that derives from a class deriving from IInterfaceTest"""
        __namespace__ = "Python.Test." + subnamespace

        def foo(self):
            return "DerivedClass"

        def base_foo(self):
            return SubClassTest.foo(self)

        def super_foo(self):
            return super(DerivedClass, self).foo()

        def bar(self, x, i):
            return "_".join([x] * i)

        def return_list(self):
            l = List[str]()
            l.Add("A")
            l.Add("B")
            l.Add("C")
            return l

    return DerivedClass

def broken_derived_class_fixture(subnamespace):
    """Delay creation of class until test starts."""

    class DerivedClass(SubClassTest):
        """class that derives from a class deriving from IInterfaceTest"""
        __namespace__ = 3

    return DerivedClass

def derived_event_test_class_fixture(subnamespace):
    """Delay creation of class until test starts."""

    class DerivedEventTest(IInterfaceTest):
        """class that implements IInterfaceTest.TestEvent"""
        __namespace__ = "Python.Test." + subnamespace

        def __init__(self):
            self.event_handlers = []

        # event handling
        def add_TestEvent(self, handler):
            self.event_handlers.append(handler)

        def remove_TestEvent(self, handler):
            self.event_handlers.remove(handler)

        def OnTestEvent(self, value):
            args = EventArgsTest(value)
            for handler in self.event_handlers:
                handler(self, args)

    return DerivedEventTest


def test_base_class():
    """Test base class managed type"""
    ob = SubClassTest()
    assert ob.foo() == "foo"
    assert FunctionsTest.test_foo(ob) == "foo"
    assert ob.bar("bar", 2) == "bar"
    assert FunctionsTest.test_bar(ob, "bar", 2) == "bar"
    assert ob.not_overriden() == "not_overriden"
    assert list(ob.return_list()) == ["a", "b", "c"]
    assert list(SubClassTest.test_list(ob)) == ["a", "b", "c"]


def test_interface():
    """Test python classes can derive from C# interfaces"""
    InterfaceTestClass = interface_test_class_fixture(test_interface.__name__)
    ob = InterfaceTestClass()
    assert ob.foo() == "InterfaceTestClass"
    assert FunctionsTest.test_foo(ob) == "InterfaceTestClass"
    assert ob.bar("bar", 2) == "bar/bar"
    assert FunctionsTest.test_bar(ob, "bar", 2) == "bar/bar"

    # pass_through will convert from InterfaceTestClass -> IInterfaceTest,
    # causing a new wrapper object to be created. Hence id will differ.
    x = FunctionsTest.pass_through_interface(ob)
    assert id(x) != id(ob)


def test_derived_class():
    """Test python class derived from managed type"""
    DerivedClass = derived_class_fixture(test_derived_class.__name__)
    ob = DerivedClass()
    assert ob.foo() == "DerivedClass"
    assert ob.base_foo() == "foo"
    assert ob.super_foo() == "foo"
    assert FunctionsTest.test_foo(ob) == "DerivedClass"
    assert ob.bar("bar", 2) == "bar_bar"
    assert FunctionsTest.test_bar(ob, "bar", 2) == "bar_bar"
    assert ob.not_overriden() == "not_overriden"
    assert list(ob.return_list()) == ["A", "B", "C"]
    assert list(SubClassTest.test_list(ob)) == ["A", "B", "C"]

    x = FunctionsTest.pass_through(ob)
    assert id(x) == id(ob)

def test_broken_derived_class():
    """Test python class derived from managed type with invalid namespace"""
    with pytest.raises(TypeError):
        DerivedClass = broken_derived_class_fixture(test_derived_class.__name__)
        ob = DerivedClass()

def test_derived_traceback():
    """Test python exception traceback in class derived from managed base"""
    class DerivedClass(SubClassTest):
        __namespace__ = "Python.Test.traceback"

        def foo(self):
            print (xyzname)
            return None

    import sys,traceback
    ob = DerivedClass()

    # direct call
    try:
        ob.foo()
        assert False
    except:
        e = sys.exc_info()
    assert "xyzname" in str(e[1])
    location = traceback.extract_tb(e[2])[-1]
    assert location[2] == "foo"

    # call through managed code
    try:
        FunctionsTest.test_foo(ob)
        assert False
    except:
        e = sys.exc_info()
    assert "xyzname" in str(e[1])
    location = traceback.extract_tb(e[2])[-1]
    assert location[2] == "foo"


def test_create_instance():
    """Test derived instances can be created from managed code"""
    DerivedClass = derived_class_fixture(test_create_instance.__name__)
    ob = FunctionsTest.create_instance(DerivedClass)
    assert ob.foo() == "DerivedClass"
    assert FunctionsTest.test_foo(ob) == "DerivedClass"
    assert ob.bar("bar", 2) == "bar_bar"
    assert FunctionsTest.test_bar(ob, "bar", 2) == "bar_bar"
    assert ob.not_overriden() == "not_overriden"

    x = FunctionsTest.pass_through(ob)
    assert id(x) == id(ob)

    InterfaceTestClass = interface_test_class_fixture(test_create_instance.__name__)
    ob2 = FunctionsTest.create_instance_interface(InterfaceTestClass)
    assert ob2.foo() == "InterfaceTestClass"
    assert FunctionsTest.test_foo(ob2) == "InterfaceTestClass"
    assert ob2.bar("bar", 2) == "bar/bar"
    assert FunctionsTest.test_bar(ob2, "bar", 2) == "bar/bar"

    y = FunctionsTest.pass_through_interface(ob2)
    assert id(y) != id(ob2)


def test_events():
    class EventHandler(object):
        def handler(self, x, args):
            self.value = args.value

    event_handler = EventHandler()

    x = SubClassTest()
    x.TestEvent += event_handler.handler
    assert FunctionsTest.test_event(x, 1) == 1
    assert event_handler.value == 1

    InterfaceTestClass = interface_test_class_fixture(test_events.__name__)
    i = InterfaceTestClass()
    with pytest.raises(System.NotImplementedException):
        FunctionsTest.test_event(i, 2)

    DerivedEventTest = derived_event_test_class_fixture(test_events.__name__)
    d = DerivedEventTest()
    d.add_TestEvent(event_handler.handler)
    assert FunctionsTest.test_event(d, 3) == 3
    assert event_handler.value == 3
    assert len(d.event_handlers) == 1


def test_isinstance_check():
    a = [str(x) for x in range(0, 1000)]
    b = [System.String(x) for x in a]

    for x in a:
        assert not isinstance(x, System.Object)
        assert not isinstance(x, System.String)

    for x in b:
        assert isinstance(x, System.Object)
        assert isinstance(x, System.String)

def test_namespace_and_init():
    calls = []
    class TestX(System.Object):
        __namespace__ = "test_clr_subclass_with_init_args"
        def __init__(self, *args, **kwargs):
            calls.append((args, kwargs))
    t = TestX(1,2,3,foo="bar")
    assert len(calls) == 1
    assert calls[0][0] == (1,2,3)
    assert calls[0][1] == {"foo":"bar"}

def test_namespace_and_argless_init():
    calls = []
    class TestX(System.Object):
        __namespace__ = "test_clr_subclass_without_init_args"
        def __init__(self):
            calls.append(True)
    t = TestX()
    assert len(calls) == 1
    assert calls[0] == True


def test_namespace_and_no_init():
    class TestX(System.Object):
        __namespace__ = "test_clr_subclass_without_init"
        q = 1
    t = TestX()
    assert t.q == 1

def test_construction_from_clr():
    import clr
    calls = []
    class TestX(System.Object):
        __namespace__ = "test_clr_subclass_init_from_clr"
        @clr.clrmethod(None, [int, str])
        def __init__(self, i, s):
            calls.append((i, s))

    # Construct a TestX from Python
    t = TestX(1, "foo")
    assert len(calls) == 1
    assert calls[0][0] == 1
    assert calls[0][1] == "foo"

    # Reset calls and construct a TestX from CLR
    calls = []
    tp = t.GetType()
    t2 = tp.GetConstructors()[0].Invoke(None)
    assert len(calls) == 0

    # The object has only been constructed, now it needs to be initialized as well
    tp.GetMethod("__init__").Invoke(t2, [1, "foo"])
    assert len(calls) == 1
    assert calls[0][0] == 1
    assert calls[0][1] == "foo"

# regression test for https://github.com/pythonnet/pythonnet/issues/1565
def test_can_be_collected_by_gc():
    from Python.Test import BaseClass

    class Derived(BaseClass):
        __namespace__ = 'test_can_be_collected_by_gc'

    inst = Derived()
    cycle = [inst]
    del inst
    cycle.append(cycle)
    del cycle

    import gc
    gc.collect()



# -*- coding: utf-8 -*-

"""Test CLR field support."""

import System
import pytest
from Python.Test import FieldTest


def test_public_instance_field():
    """Test public instance fields."""
    ob = FieldTest()
    assert ob.PublicField == 0

    ob.PublicField = 1
    assert ob.PublicField == 1

    with pytest.raises(TypeError):
        del FieldTest().PublicField


def test_public_static_field():
    """Test public static fields."""
    ob = FieldTest()
    assert FieldTest.PublicStaticField == 0

    FieldTest.PublicStaticField = 1
    assert FieldTest.PublicStaticField == 1

    assert ob.PublicStaticField == 1
    ob.PublicStaticField = 0
    assert ob.PublicStaticField == 0

    with pytest.raises(TypeError):
        del FieldTest.PublicStaticField

    with pytest.raises(TypeError):
        del FieldTest().PublicStaticField


def test_protected_instance_field():
    """Test protected instance fields."""
    ob = FieldTest()
    assert ob.ProtectedField == 0

    ob.ProtectedField = 1
    assert ob.ProtectedField == 1

    with pytest.raises(TypeError):
        del FieldTest().ProtectedField


def test_protected_static_field():
    """Test protected static fields."""
    ob = FieldTest()
    assert FieldTest.ProtectedStaticField == 0

    FieldTest.ProtectedStaticField = 1
    assert FieldTest.ProtectedStaticField == 1

    assert ob.ProtectedStaticField == 1
    ob.ProtectedStaticField = 0
    assert ob.ProtectedStaticField == 0

    with pytest.raises(TypeError):
        del FieldTest.ProtectedStaticField

    with pytest.raises(TypeError):
        del FieldTest().ProtectedStaticField


def test_read_only_instance_field():
    """Test readonly instance fields."""
    assert FieldTest().ReadOnlyField == 0

    with pytest.raises(TypeError):
        FieldTest().ReadOnlyField = 1

    with pytest.raises(TypeError):
        del FieldTest().ReadOnlyField


def test_read_only_static_field():
    """Test readonly static fields."""
    ob = FieldTest()

    assert FieldTest.ReadOnlyStaticField == 0
    assert ob.ReadOnlyStaticField == 0

    with pytest.raises(TypeError):
        FieldTest.ReadOnlyStaticField = 1

    with pytest.raises(TypeError):
        FieldTest().ReadOnlyStaticField = 1

    with pytest.raises(TypeError):
        del FieldTest.ReadOnlyStaticField

    with pytest.raises(TypeError):
        del FieldTest().ReadOnlyStaticField


def test_constant_field():
    """Test const fields."""
    ob = FieldTest()

    assert FieldTest.ConstField == 0
    assert ob.ConstField == 0

    with pytest.raises(TypeError):
        FieldTest().ConstField = 1

    with pytest.raises(TypeError):
        FieldTest.ConstField = 1

    with pytest.raises(TypeError):
        del FieldTest().ConstField

    with pytest.raises(TypeError):
        del FieldTest.ConstField


def test_internal_field():
    """Test internal fields."""

    with pytest.raises(AttributeError):
        _ = FieldTest().InternalField

    with pytest.raises(AttributeError):
        _ = FieldTest().InternalStaticField

    with pytest.raises(AttributeError):
        _ = FieldTest.InternalStaticField


def test_private_field():
    """Test private fields."""

    with pytest.raises(AttributeError):
        _ = FieldTest().PrivateField

    with pytest.raises(AttributeError):
        _ = FieldTest().PrivateStaticField

    with pytest.raises(AttributeError):
        _ = FieldTest.PrivateStaticField


def test_field_descriptor_get_set():
    """Test field descriptor get / set."""

    # This test ensures that setting an attribute implemented with
    # a descriptor actually goes through the descriptor (rather than
    # silently replacing the descriptor in the instance or type dict.

    ob = FieldTest()

    assert FieldTest.PublicStaticField == 0
    assert ob.PublicStaticField == 0

    descriptor = FieldTest.__dict__['PublicStaticField']
    assert type(descriptor) != int

    ob.PublicStaticField = 0
    descriptor = FieldTest.__dict__['PublicStaticField']
    assert type(descriptor) != int

    FieldTest.PublicStaticField = 0
    descriptor = FieldTest.__dict__['PublicStaticField']
    assert type(descriptor) != int


def test_field_descriptor_wrong_type():
    """Test setting a field using a value of the wrong type."""

    with pytest.raises(TypeError):
        FieldTest().PublicField = "spam"


def test_field_descriptor_abuse():
    """Test field descriptor abuse."""
    desc = FieldTest.__dict__['PublicField']

    with pytest.raises(TypeError):
        desc.__get__(0, 0)

    with pytest.raises(TypeError):
        desc.__set__(0, 0)


def test_boolean_field():
    """Test boolean fields."""
    # change this to true / false later for Python 2.3?
    ob = FieldTest()
    assert ob.BooleanField is False

    ob.BooleanField = True
    assert ob.BooleanField is True

    ob.BooleanField = False
    assert ob.BooleanField is False

    with pytest.raises(TypeError):
        ob.BooleanField = 1

    with pytest.raises(TypeError):
        ob.BooleanField = 0


def test_sbyte_field():
    """Test sbyte fields."""
    ob = FieldTest()
    assert ob.SByteField == 0

    ob.SByteField = 1
    assert ob.SByteField == 1


def test_byte_field():
    """Test byte fields."""
    ob = FieldTest()
    assert ob.ByteField == 0

    ob.ByteField = 1
    assert ob.ByteField == 1


def test_char_field():
    """Test char fields."""
    ob = FieldTest()
    assert ob.CharField == u'A'
    assert ob.CharField == 'A'

    ob.CharField = 'B'
    assert ob.CharField == u'B'
    assert ob.CharField == 'B'

    ob.CharField = u'C'
    assert ob.CharField == u'C'
    assert ob.CharField == 'C'


def test_int16_field():
    """Test int16 fields."""
    ob = FieldTest()
    assert ob.Int16Field == 0

    ob.Int16Field = 1
    assert ob.Int16Field == 1


def test_int32_field():
    """Test int32 fields."""
    ob = FieldTest()
    assert ob.Int32Field == 0

    ob.Int32Field = 1
    assert ob.Int32Field == 1


def test_int64_field():
    """Test int64 fields."""
    ob = FieldTest()
    assert ob.Int64Field == 0

    ob.Int64Field = 1
    assert ob.Int64Field == 1


def test_uint16_field():
    """Test uint16 fields."""
    ob = FieldTest()
    assert ob.UInt16Field == 0

    ob.UInt16Field = 1
    assert ob.UInt16Field == 1


def test_uint32_field():
    """Test uint32 fields."""
    ob = FieldTest()
    assert ob.UInt32Field == 0

    ob.UInt32Field = 1
    assert ob.UInt32Field == 1


def test_uint64_field():
    """Test uint64 fields."""
    ob = FieldTest()
    assert ob.UInt64Field == 0

    ob.UInt64Field = 1
    assert ob.UInt64Field == 1


def test_single_field():
    """Test single fields."""
    ob = FieldTest()
    assert ob.SingleField == 0.0

    ob.SingleField = 1.1
    assert ob.SingleField == System.Single(1.1)


def test_double_field():
    """Test double fields."""
    ob = FieldTest()
    assert ob.DoubleField == 0.0

    ob.DoubleField = 1.1
    assert ob.DoubleField == 1.1


def test_decimal_field():
    """Test decimal fields."""
    ob = FieldTest()
    assert ob.DecimalField == System.Decimal(0)

    ob.DecimalField = System.Decimal(1)
    assert ob.DecimalField == System.Decimal(1)


def test_string_field():
    """Test string fields."""
    ob = FieldTest()
    assert ob.StringField == "spam"

    ob.StringField = "eggs"
    assert ob.StringField == "eggs"


def test_interface_field():
    """Test interface fields."""
    from Python.Test import Spam, ISpam

    ob = FieldTest()

    assert ISpam(ob.SpamField).GetValue() == "spam"
    assert ob.SpamField.GetValue() == "spam"

    ob.SpamField = Spam("eggs")
    assert ISpam(ob.SpamField).GetValue() == "eggs"
    assert ob.SpamField.GetValue() == "eggs"


def test_object_field():
    """Test ob fields."""
    ob = FieldTest()
    assert ob.ObjectField is None

    ob.ObjectField = System.String("spam")
    assert ob.ObjectField == "spam"

    ob.ObjectField = System.Int32(1)
    assert ob.ObjectField == 1

    ob.ObjectField = None
    assert ob.ObjectField is None


def test_enum_field():
    """Test enum fields."""
    from Python.Test import ShortEnum

    ob = FieldTest()
    assert ob.EnumField == ShortEnum.Zero

    ob.EnumField = ShortEnum.One
    assert ob.EnumField == ShortEnum.One


def test_nullable_field():
    """Test nullable fields."""
    ob = FieldTest()

    ob.StringField = None
    assert ob.StringField is None

    ob.ObjectField = None
    assert ob.ObjectField is None

    ob.SpamField = None
    assert ob.SpamField is None

    # Primitive types and enums should not be set to null.

    with pytest.raises(TypeError):
        FieldTest().Int32Field = None

    with pytest.raises(TypeError):
        FieldTest().EnumField = None



"""Test CLR <-> Python type conversions."""

import operator
import pytest

import System
from Python.Test import ConversionTest, MethodResolutionInt, UnicodeString
from Python.Runtime import PyObjectConversions
from Python.Runtime.Codecs import RawProxyEncoder


def test_bool_conversion():
    """Test bool conversion."""
    ob = ConversionTest()
    assert ob.BooleanField is False
    assert ob.BooleanField == 0

    ob.BooleanField = True
    assert ob.BooleanField is True
    assert ob.BooleanField == 1

    ob.BooleanField = False
    assert ob.BooleanField is False
    assert ob.BooleanField == 0

    with pytest.raises(TypeError):
        ob.BooleanField = 1
    
    with pytest.raises(TypeError):
        ob.BooleanField = 0

    with pytest.raises(TypeError):
        ob.BooleanField = None

    with pytest.raises(TypeError):
        ob.BooleanField = ''

    with pytest.raises(TypeError):
        ob.BooleanField = System.Boolean(0)

    with pytest.raises(TypeError):
        ob.BooleanField = System.Boolean(1)

    with pytest.raises(TypeError):
        ob.BooleanField = System.Boolean('a')


def test_sbyte_conversion():
    """Test sbyte conversion."""
    assert System.SByte.MaxValue == 127
    assert System.SByte.MinValue == -128

    ob = ConversionTest()
    assert ob.SByteField == 0

    ob.SByteField = 127
    assert ob.SByteField == 127

    ob.SByteField = -128
    assert ob.SByteField == -128

    ob.SByteField = System.SByte(127)
    assert ob.SByteField == 127

    ob.SByteField = System.SByte(-128)
    assert ob.SByteField == -128

    with pytest.raises(TypeError):
        ConversionTest().SByteField = "spam"

    with pytest.raises(TypeError):
        ConversionTest().SByteField = None

    with pytest.raises(OverflowError):
        ConversionTest().SByteField = 128

    with pytest.raises(OverflowError):
        ConversionTest().SByteField = -129

    with pytest.raises(OverflowError):
        _ = System.SByte(128)

    with pytest.raises(OverflowError):
        _ = System.SByte(-129)


def test_byte_conversion():
    """Test byte conversion."""
    assert System.Byte.MaxValue == 255
    assert System.Byte.MinValue == 0

    ob = ConversionTest()
    assert ob.ByteField == 0

    ob.ByteField = 255
    assert ob.ByteField == 255

    ob.ByteField = 0
    assert ob.ByteField == 0

    ob.ByteField = System.Byte(255)
    assert ob.ByteField == 255

    ob.ByteField = System.Byte(0)
    assert ob.ByteField == 0

    with pytest.raises(TypeError):
        ConversionTest().ByteField = "spam"

    with pytest.raises(TypeError):
        ConversionTest().ByteField = None

    with pytest.raises(OverflowError):
        ConversionTest().ByteField = 256

    with pytest.raises(OverflowError):
        ConversionTest().ByteField = -1

    with pytest.raises(OverflowError):
        _ = System.Byte(256)

    with pytest.raises(OverflowError):
        _ = System.Byte(-1)


def test_char_conversion():
    """Test char conversion."""
    assert System.Char.MaxValue == chr(65535)
    assert System.Char.MinValue == chr(0)

    ob = ConversionTest()
    assert ob.CharField == u'A'

    ob.CharField = 'B'
    assert ob.CharField == u'B'

    ob.CharField = u'B'
    assert ob.CharField == u'B'

    ob.CharField = 67
    assert ob.CharField == u'C'

    with pytest.raises(OverflowError):
        ConversionTest().CharField = 65536

    with pytest.raises(OverflowError):
        ConversionTest().CharField = -1

    with pytest.raises(TypeError):
        ConversionTest().CharField = None


def test_int16_conversion():
    """Test int16 conversion."""
    assert System.Int16.MaxValue == 32767
    assert System.Int16.MinValue == -32768

    ob = ConversionTest()
    assert ob.Int16Field == 0

    ob.Int16Field = 32767
    assert ob.Int16Field == 32767

    ob.Int16Field = -32768
    assert ob.Int16Field == -32768

    ob.Int16Field = System.Int16(32767)
    assert ob.Int16Field == 32767

    ob.Int16Field = System.Int16(-32768)
    assert ob.Int16Field == -32768

    with pytest.raises(TypeError):
        ConversionTest().Int16Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().Int16Field = None

    with pytest.raises(OverflowError):
        ConversionTest().Int16Field = 32768

    with pytest.raises(OverflowError):
        ConversionTest().Int16Field = -32769

    with pytest.raises(OverflowError):
        _ = System.Int16(32768)

    with pytest.raises(OverflowError):
        _ = System.Int16(-32769)


def test_int32_conversion():
    """Test int32 conversion."""
    assert System.Int32.MaxValue == 2147483647
    assert System.Int32.MinValue == -2147483648

    ob = ConversionTest()
    assert ob.Int32Field == 0

    ob.Int32Field = 2147483647
    assert ob.Int32Field == 2147483647

    ob.Int32Field = -2147483648
    assert ob.Int32Field == -2147483648

    ob.Int32Field = System.Int32(2147483647)
    assert ob.Int32Field == 2147483647

    ob.Int32Field = System.Int32(-2147483648)
    assert ob.Int32Field == -2147483648

    with pytest.raises(TypeError):
        ConversionTest().Int32Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().Int32Field = None

    with pytest.raises(OverflowError):
        ConversionTest().Int32Field = 2147483648

    with pytest.raises(OverflowError):
        ConversionTest().Int32Field = -2147483649

    with pytest.raises(OverflowError):
        _ = System.Int32(2147483648)

    with pytest.raises(OverflowError):
        _ = System.Int32(-2147483649)


def test_int64_conversion():
    """Test int64 conversion."""
    assert System.Int64.MaxValue == 9223372036854775807
    assert System.Int64.MinValue == -9223372036854775808

    ob = ConversionTest()
    assert ob.Int64Field == 0

    ob.Int64Field = 9223372036854775807
    assert ob.Int64Field == 9223372036854775807

    ob.Int64Field = -9223372036854775808
    assert ob.Int64Field == -9223372036854775808

    ob.Int64Field = System.Int64(9223372036854775807)
    assert ob.Int64Field == 9223372036854775807

    ob.Int64Field = System.Int64(-9223372036854775808)
    assert ob.Int64Field == -9223372036854775808

    with pytest.raises(TypeError):
        ConversionTest().Int64Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().Int64Field = None

    with pytest.raises(OverflowError):
        ConversionTest().Int64Field = 9223372036854775808

    with pytest.raises(OverflowError):
        ConversionTest().Int64Field = -9223372036854775809

    with pytest.raises(OverflowError):
        _ = System.Int64(9223372036854775808)

    with pytest.raises(OverflowError):
        _ = System.Int64(-9223372036854775809)


def test_uint16_conversion():
    """Test uint16 conversion."""
    assert System.UInt16.MaxValue == 65535
    assert System.UInt16.MinValue == 0

    ob = ConversionTest()
    assert ob.UInt16Field == 0

    ob.UInt16Field = 65535
    assert ob.UInt16Field == 65535

    ob.UInt16Field = -0
    assert ob.UInt16Field == 0

    ob.UInt16Field = System.UInt16(65535)
    assert ob.UInt16Field == 65535

    ob.UInt16Field = System.UInt16(0)
    assert ob.UInt16Field == 0

    with pytest.raises(TypeError):
        ConversionTest().UInt16Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().UInt16Field = None

    with pytest.raises(OverflowError):
        ConversionTest().UInt16Field = 65536

    with pytest.raises(OverflowError):
        ConversionTest().UInt16Field = -1

    with pytest.raises(OverflowError):
        _ = System.UInt16(65536)

    with pytest.raises(OverflowError):
        _ = System.UInt16(-1)


def test_uint32_conversion():
    """Test uint32 conversion."""
    assert System.UInt32.MaxValue == 4294967295
    assert System.UInt32.MinValue == 0

    ob = ConversionTest()
    assert ob.UInt32Field == 0

    ob.UInt32Field = 4294967295
    assert ob.UInt32Field == 4294967295

    ob.UInt32Field = -0
    assert ob.UInt32Field == 0

    ob.UInt32Field = System.UInt32(4294967295)
    assert ob.UInt32Field == 4294967295

    ob.UInt32Field = System.UInt32(0)
    assert ob.UInt32Field == 0

    with pytest.raises(TypeError):
        ConversionTest().UInt32Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().UInt32Field = None

    with pytest.raises(OverflowError):
        ConversionTest().UInt32Field = 4294967296

    with pytest.raises(OverflowError):
        ConversionTest().UInt32Field = -1

    with pytest.raises(OverflowError):
        _ = System.UInt32(4294967296)

    with pytest.raises(OverflowError):
        _ = System.UInt32(-1)


def test_uint64_conversion():
    """Test uint64 conversion."""
    assert System.UInt64.MaxValue == 18446744073709551615
    assert System.UInt64.MinValue == 0

    ob = ConversionTest()
    assert ob.UInt64Field == 0

    ob.UInt64Field = 18446744073709551615
    assert ob.UInt64Field == 18446744073709551615

    ob.UInt64Field = -0
    assert ob.UInt64Field == 0

    ob.UInt64Field = System.UInt64(18446744073709551615)
    assert ob.UInt64Field == 18446744073709551615

    ob.UInt64Field = System.UInt64(0)
    assert ob.UInt64Field == 0

    with pytest.raises(TypeError):
        ConversionTest().UInt64Field = 0.5

    with pytest.raises(TypeError):
        ConversionTest().UInt64Field = "spam"

    with pytest.raises(TypeError):
        ConversionTest().UInt64Field = None

    with pytest.raises(OverflowError):
        ConversionTest().UInt64Field = 18446744073709551616

    with pytest.raises(OverflowError):
        ConversionTest().UInt64Field = -1

    with pytest.raises(OverflowError):
        _ = System.UInt64((18446744073709551616))

    with pytest.raises(OverflowError):
        _ = System.UInt64(-1)


def test_single_conversion():
    """Test single conversion."""
    assert System.Single.MaxValue == pytest.approx(3.402823e38)
    assert System.Single.MinValue == pytest.approx(-3.402823e38)

    ob = ConversionTest()
    assert ob.SingleField == 0.0

    ob.SingleField = 3.402823e38
    assert ob.SingleField == System.Single(3.402823e38)

    ob.SingleField = -3.402823e38
    assert ob.SingleField == System.Single(-3.402823e38)

    with pytest.raises(TypeError):
        ConversionTest().SingleField = "spam"

    with pytest.raises(TypeError):
        ConversionTest().SingleField = None

    with pytest.raises(OverflowError):
        ConversionTest().SingleField = 3.402824e38

    with pytest.raises(OverflowError):
        ConversionTest().SingleField = -3.402824e38

    with pytest.raises(OverflowError):
        _ = System.Single(3.402824e38)

    with pytest.raises(OverflowError):
        _ = System.Single(-3.402824e38)


def test_double_conversion():
    """Test double conversion."""
    assert System.Double.MaxValue == 1.7976931348623157e308
    assert System.Double.MinValue == -1.7976931348623157e308

    ob = ConversionTest()
    assert ob.DoubleField == 0.0

    ob.DoubleField = 1.7976931348623157e308
    assert ob.DoubleField == 1.7976931348623157e308

    ob.DoubleField = -1.7976931348623157e308
    assert ob.DoubleField == -1.7976931348623157e308

    ob.DoubleField = System.Double(1.7976931348623157e308)
    assert ob.DoubleField == 1.7976931348623157e308

    ob.DoubleField = System.Double(-1.7976931348623157e308)
    assert ob.DoubleField == -1.7976931348623157e308

    with pytest.raises(TypeError):
        ConversionTest().DoubleField = "spam"

    with pytest.raises(TypeError):
        ConversionTest().DoubleField = None



def test_decimal_conversion():
    """Test decimal conversion."""
    from System import Decimal

    max_d = Decimal.Parse("79228162514264337593543950335")
    min_d = Decimal.Parse("-79228162514264337593543950335")

    assert Decimal.ToInt64(Decimal(10)) == 10

    ob = ConversionTest()
    assert ob.DecimalField == Decimal(0)

    ob.DecimalField = Decimal(10)
    assert ob.DecimalField == Decimal(10)

    ob.DecimalField = Decimal.One
    assert ob.DecimalField == Decimal.One

    ob.DecimalField = Decimal.Zero
    assert ob.DecimalField == Decimal.Zero

    ob.DecimalField = max_d
    assert ob.DecimalField == max_d

    ob.DecimalField = min_d
    assert ob.DecimalField == min_d

    with pytest.raises(TypeError):
        ConversionTest().DecimalField = None

    with pytest.raises(TypeError):
        ConversionTest().DecimalField = "spam"

    with pytest.raises(TypeError):
        ConversionTest().DecimalField = 1


def test_string_conversion():
    """Test string / unicode conversion."""
    ob = ConversionTest()

    assert ob.StringField == "spam"
    assert ob.StringField == u"spam"

    ob.StringField = "eggs"
    assert ob.StringField == "eggs"
    assert ob.StringField == u"eggs"

    ob.StringField = u"spam"
    assert ob.StringField == "spam"
    assert ob.StringField == u"spam"

    ob.StringField = u'\uffff\uffff'
    assert ob.StringField == u'\uffff\uffff'

    ob.StringField = System.String("spam")
    assert ob.StringField == "spam"
    assert ob.StringField == u"spam"

    ob.StringField = System.String(u'\uffff\uffff')
    assert ob.StringField == u'\uffff\uffff'

    ob.StringField = None
    assert ob.StringField is None

    with pytest.raises(TypeError):
        ConversionTest().StringField = 1

    world = UnicodeString()
    test_unicode_str = u"안녕"
    assert test_unicode_str == str(world.value)
    assert test_unicode_str == str(world.GetString())
    assert test_unicode_str == str(world)


def test_interface_conversion():
    """Test interface conversion."""
    from Python.Test import Spam, ISpam

    ob = ConversionTest()

    assert ISpam(ob.SpamField).GetValue() == "spam"
    assert ob.SpamField.GetValue() == "spam"

    ob.SpamField = Spam("eggs")
    assert ISpam(ob.SpamField).GetValue() == "eggs"
    assert ob.SpamField.GetValue() == "eggs"

    # need to test spam subclass here.

    ob.SpamField = None
    assert ob.SpamField is None

    with pytest.raises(TypeError):
        ob = ConversionTest()
        ob.SpamField = System.String("bad")

    with pytest.raises(TypeError):
        ob = ConversionTest()
        ob.SpamField = System.Int32(1)


def test_object_conversion():
    """Test ob conversion."""
    from Python.Test import Spam

    ob = ConversionTest()
    assert ob.ObjectField is None

    ob.ObjectField = Spam("eggs")
    assert ob.ObjectField.__class__.__name__ == "Spam"
    assert ob.ObjectField.GetValue() == "eggs"

    ob.ObjectField = None
    assert ob.ObjectField is None

    ob.ObjectField = System.String("spam")
    assert ob.ObjectField == "spam"

    ob.ObjectField = System.Int32(1)
    assert ob.ObjectField == 1

    # need to test subclass here

    class Foo(object):
        pass
    ob.ObjectField = Foo
    assert ob.ObjectField == Foo


def test_null_conversion():
    """Test null conversion."""
    import System

    ob = ConversionTest()

    ob.StringField = None
    assert ob.StringField is None

    ob.ObjectField = None
    assert ob.ObjectField is None

    ob.SpamField = None
    assert ob.SpamField is None

    pi = 22/7
    assert ob.Echo[System.Double](pi) == pi
    assert ob.Echo[System.DateTime](None) is None

    # Primitive types and enums should not be set to null.

    with pytest.raises(TypeError):
        ConversionTest().Int32Field = None

    with pytest.raises(TypeError):
        ConversionTest().EnumField = None


def test_byte_array_conversion():
    """Test byte array conversion."""
    ob = ConversionTest()

    assert ob.ByteArrayField is None

    ob.ByteArrayField = [0, 1, 2, 3, 4]
    array = ob.ByteArrayField
    assert len(array) == 5
    assert array[0] == 0
    assert array[4] == 4

    value = b"testing"
    ob.ByteArrayField = value
    array = ob.ByteArrayField
    for i, _ in enumerate(value):
        assert array[i] == operator.getitem(value, i)


def test_sbyte_array_conversion():
    """Test sbyte array conversion."""
    ob = ConversionTest()

    assert ob.SByteArrayField is None

    ob.SByteArrayField = [0, 1, 2, 3, 4]
    array = ob.SByteArrayField
    assert len(array) == 5
    assert array[0] == 0
    assert array[4] == 4

    value = b"testing"
    ob.SByteArrayField = value
    array = ob.SByteArrayField
    for i, _ in enumerate(value):
        assert array[i] == operator.getitem(value, i)

def test_codecs():
    """Test codec registration from Python"""
    class ListAsRawEncoder(RawProxyEncoder):
        __namespace__ = "Python.Test"
        def CanEncode(self, clr_type):
            return clr_type.Name == "List`1" and clr_type.Namespace == "System.Collections.Generic"

    list_raw_encoder = ListAsRawEncoder()
    PyObjectConversions.RegisterEncoder(list_raw_encoder)

    ob = ConversionTest()

    l = ob.ListField
    l.Add(42)
    assert ob.ListField.Count == 1

def test_int_param_resolution_required():
    """Test resolution of `int` parameters when resolution is needed"""

    mri = MethodResolutionInt()
    data = list(mri.MethodA(0x1000, 10))
    assert len(data) == 10
    assert data[0] == 0

    data = list(mri.MethodA(0x100000000, 10))
    assert len(data) == 10
    assert data[0] == 0



"""
Implements collections.abc for common .NET types
https://docs.python.org/3.6/library/collections.abc.html
"""

import collections.abc as col

class IteratorMixin(col.Iterator):
    def close(self):
        self.Dispose()

class IterableMixin(col.Iterable):
    pass

class SizedMixin(col.Sized):
    def __len__(self): return self.Count

class ContainerMixin(col.Container):
    def __contains__(self, item): return self.Contains(item)

try:
    abc_Collection = col.Collection
except AttributeError:
    # Python 3.5- does not have collections.abc.Collection
    abc_Collection = col.Container

class CollectionMixin(SizedMixin, IterableMixin, ContainerMixin, abc_Collection):
    pass

class SequenceMixin(CollectionMixin, col.Sequence):
    pass

class MutableSequenceMixin(SequenceMixin, col.MutableSequence):
    pass

class MappingMixin(CollectionMixin, col.Mapping):
    def __contains__(self, item): return self.ContainsKey(item)
    def keys(self): return self.Keys
    def items(self): return [(k,self.get(k)) for k in self.Keys]
    def values(self): return self.Values
    def __iter__(self): return self.Keys.__iter__()
    def get(self, key, default=None):
        existed, item = self.TryGetValue(key, None)
        return item if existed else default

class MutableMappingMixin(MappingMixin, col.MutableMapping):
    _UNSET_ = object()

    def __delitem__(self, key):
        self.Remove(key)

    def clear(self):
        self.Clear()

    def pop(self, key, default=_UNSET_):
        existed, item = self.TryGetValue(key, None)
        if existed:
            self.Remove(key)
            return item
        elif default == self._UNSET_:
            raise KeyError(key)
        else:
            return default

    def setdefault(self, key, value=None):
        existed, item = self.TryGetValue(key, None)
        if existed:
            return item
        else:
            self[key] = value
            return value

    def update(self, items, **kwargs):
        if isinstance(items, col.Mapping):
            for key, value in items.items():
                self[key] = value
        else:
            for key, value in items:
                self[key] = value

        for key, value in kwargs.items():
            self[key] = value



# -*- coding: utf-8 -*-

import sys
try:
    del sys.modules["System.IO"]
except KeyError:
    pass

assert "FileStream" not in globals()
import System.IO
from System.IO import *

assert "FileStream" in globals()



import sys
import clr_loader

_RUNTIME = None
_LOADER_ASSEMBLY = None
_FFI = None
_LOADED = False


def set_runtime(runtime):
    global _RUNTIME
    if _LOADED:
        raise RuntimeError("The runtime {} has already been loaded".format(_RUNTIME))

    _RUNTIME = runtime


def set_default_runtime() -> None:
    if sys.platform == 'win32':
        set_runtime(clr_loader.get_netfx())
    else:
        set_runtime(clr_loader.get_mono())


def load():
    global _FFI, _LOADED, _LOADER_ASSEMBLY

    if _LOADED:
        return

    from os.path import join, dirname

    if _RUNTIME is None:
        # TODO: Warn, in the future the runtime must be set explicitly, either
        # as a config/env variable or via set_runtime
        set_default_runtime()

    dll_path = join(dirname(__file__), "runtime", "Python.Runtime.dll")
    
    _LOADER_ASSEMBLY = _RUNTIME.get_assembly(dll_path)

    func = _LOADER_ASSEMBLY["Python.Runtime.Loader.Initialize"]
    if func(''.encode("utf8")) != 0:
        raise RuntimeError("Failed to initialize Python.Runtime.dll")

    import atexit
    atexit.register(unload)


def unload():
    global _RUNTIME
    if _LOADER_ASSEMBLY is not None:
        func = _LOADER_ASSEMBLY["Python.Runtime.Loader.Shutdown"]
        if func(b"") != 0:
            raise RuntimeError("Failed to call Python.NET shutdown")

    if _RUNTIME is not None:
        # TODO: Add explicit `close` to clr_loader
        _RUNTIME = None



# -*- coding: utf-8 -*-

"""Helper script to test argv.
Ensures that argv isn't modified after importing clr.
For more details see GH#404 - argv not found"""

from __future__ import print_function

import sys
import clr

print(sys.argv)



# -*- coding: utf-8 -*-

"""Test CLR event support."""

import pytest
from Python.Test import EventTest, EventArgsTest

from .utils import (CallableHandler, ClassMethodHandler, GenericHandler,
                    MultipleHandler, StaticMethodHandler, VarCallableHandler,
                    VariableArgsHandler)


def test_public_instance_event():
    """Test public instance events."""
    ob = EventTest()

    handler = GenericHandler()
    assert handler.value is None

    ob.PublicEvent += handler.handler

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler.handler


def test_public_static_event():
    """Test public static events."""
    handler = GenericHandler()
    assert handler.value is None

    EventTest.PublicStaticEvent += handler.handler

    EventTest.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 10


def test_protected_instance_event():
    """Test protected instance events."""
    ob = EventTest()

    handler = GenericHandler()
    assert handler.value is None

    ob.ProtectedEvent += handler.handler

    ob.OnProtectedEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.ProtectedEvent -= handler.handler


def test_protected_static_event():
    """Test protected static events."""
    handler = GenericHandler()
    assert handler.value is None

    EventTest.ProtectedStaticEvent += handler.handler

    EventTest.OnProtectedStaticEvent(EventArgsTest(10))
    assert handler.value == 10

    EventTest.ProtectedStaticEvent -= handler.handler


def test_internal_events():
    """Test internal events."""

    with pytest.raises(AttributeError):
        _ = EventTest().InternalEvent

    with pytest.raises(AttributeError):
        _ = EventTest().InternalStaticEvent

    with pytest.raises(AttributeError):
        _ = EventTest.InternalStaticEvent


def test_private_events():
    """Test private events."""

    with pytest.raises(AttributeError):
        _ = EventTest().PrivateEvent

    with pytest.raises(AttributeError):
        _ = EventTest().PrivateStaticEvent

    with pytest.raises(AttributeError):
        _ = EventTest.PrivateStaticEvent


def test_multicast_event():
    """Test multicast events."""
    ob = EventTest()

    handler1 = GenericHandler()
    handler2 = GenericHandler()
    handler3 = GenericHandler()

    ob.PublicEvent += handler1.handler
    ob.PublicEvent += handler2.handler
    ob.PublicEvent += handler3.handler

    ob.OnPublicEvent(EventArgsTest(10))

    assert handler1.value == 10
    assert handler2.value == 10
    assert handler3.value == 10

    ob.OnPublicEvent(EventArgsTest(20))

    assert handler1.value == 20
    assert handler2.value == 20
    assert handler3.value == 20

    ob.PublicEvent -= handler1.handler
    ob.PublicEvent -= handler2.handler
    ob.PublicEvent -= handler3.handler


def test_instance_method_handler():
    """Test instance method handlers."""
    ob = EventTest()
    handler = GenericHandler()

    ob.PublicEvent += handler.handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler.handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_var_args_instance_method_handler():
    """Test vararg instance method handlers."""
    ob = EventTest()
    handler = VariableArgsHandler()

    ob.PublicEvent += handler.handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler.handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_callableob_handler():
    """Test callable ob handlers."""
    ob = EventTest()
    handler = CallableHandler()

    ob.PublicEvent += handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_var_args_callable_handler():
    """Test varargs callable handlers."""
    ob = EventTest()
    handler = VarCallableHandler()

    ob.PublicEvent += handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_static_method_handler():
    """Test static method handlers."""
    ob = EventTest()
    handler = StaticMethodHandler()
    StaticMethodHandler.value = None

    ob.PublicEvent += handler.handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler.handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_class_method_handler():
    """Test class method handlers."""
    ob = EventTest()
    handler = ClassMethodHandler()
    ClassMethodHandler.value = None

    ob.PublicEvent += handler.handler
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.PublicEvent -= handler.handler
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_managed_instance_method_handler():
    """Test managed instance method handlers."""
    ob = EventTest()

    ob.PublicEvent += ob.GenericHandler
    assert ob.value == 0

    ob.OnPublicEvent(EventArgsTest(10))
    assert ob.value == 10

    ob.PublicEvent -= ob.GenericHandler
    assert ob.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert ob.value == 10


def test_managed_static_method_handler():
    """Test managed static method handlers."""
    ob = EventTest()
    EventTest.s_value = 0

    ob.PublicEvent += ob.StaticHandler
    assert EventTest.s_value == 0

    ob.OnPublicEvent(EventArgsTest(10))
    assert EventTest.s_value == 10

    ob.PublicEvent -= ob.StaticHandler
    assert EventTest.s_value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert EventTest.s_value == 10


def test_unbound_method_handler():
    """Test failure mode for unbound method handlers."""
    ob = EventTest()
    ob.PublicEvent += GenericHandler.handler

    with pytest.raises(TypeError):
        ob.OnPublicEvent(EventArgsTest(10))

    ob.PublicEvent -= GenericHandler.handler


def test_function_handler():
    """Test function handlers."""
    ob = EventTest()
    dict_ = {'value': None}

    def handler(sender, args, dict_=dict_):
        dict_['value'] = args.value

    ob.PublicEvent += handler
    assert dict_['value'] is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert dict_['value'] == 10

    ob.PublicEvent -= handler
    assert dict_['value'] == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert dict_['value'] == 10

def test_out_function_handler():
    """Test function handlers with Out arguments."""
    ob = EventTest()

    value = 10
    def handler(ignored):
        return value

    ob.OutIntEvent += handler
    result = ob.OnOutIntEvent(55)
    assert result == value

    ob.OutStringEvent += handler
    value = 'This is the event data'
    result = ob.OnOutStringEvent('Hello')
    assert result == value

def test_ref_function_handler():
    """Test function handlers with Ref arguments."""
    ob = EventTest()

    value = 10
    def handler(data):
        return value + data

    ob.RefIntEvent += ob.RefIntHandler
    ob.RefIntEvent += handler
    result = ob.OnRefIntEvent(5)
    assert result == value + 5 + 1

    ob.RefStringEvent += ob.RefStringHandler
    ob.RefStringEvent += handler
    value = 'This is the event data'
    result = ob.OnRefStringEvent('!')
    assert result == value + '!!'

def test_add_non_callable_handler():
    """Test handling of attempts to add non-callable handlers."""

    with pytest.raises(TypeError):
        ob = EventTest()
        ob.PublicEvent += 10

    with pytest.raises(TypeError):
        ob = EventTest()
        ob.PublicEvent += "spam"

    with pytest.raises(TypeError):
        class Spam(object):
            pass

        ob = EventTest()
        ob.PublicEvent += Spam()


def test_remove_multiple_handlers():
    """Test removing multiple instances of the same handler."""
    ob = EventTest()
    handler = MultipleHandler()

    h1 = handler.handler
    ob.PublicEvent += h1

    h2 = handler.handler
    ob.PublicEvent += h2

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 20

    ob.PublicEvent -= h1

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 30

    ob.PublicEvent -= h2

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 30

    # try again, removing in a different order.

    ob = EventTest()
    handler = MultipleHandler()

    h1 = handler.handler
    ob.PublicEvent += h1

    h2 = handler.handler
    ob.PublicEvent += h2

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 20

    ob.PublicEvent -= h2

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 30

    ob.PublicEvent -= h1

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 30


def test_remove_multiple_static_handlers():
    """Test removing multiple instances of a static handler."""
    ob = EventTest()
    handler = MultipleHandler()

    h1 = handler.handler
    ob.PublicStaticEvent += h1

    h2 = handler.handler
    ob.PublicStaticEvent += h2

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 20

    ob.PublicStaticEvent -= h1

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 30

    ob.PublicStaticEvent -= h2

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 30

    # try again, removing in a different order.

    ob = EventTest()
    handler = MultipleHandler()

    h1 = handler.handler
    ob.PublicStaticEvent += h1

    h2 = handler.handler
    ob.PublicStaticEvent += h2

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 20

    ob.PublicStaticEvent -= h2

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 30

    ob.PublicStaticEvent -= h1

    ob.OnPublicStaticEvent(EventArgsTest(10))
    assert handler.value == 30


def test_random_multiple_handlers():
    """Test random subscribe / unsubscribe of the same handlers."""
    import random
    ob = EventTest()
    handler = MultipleHandler()
    handler2 = MultipleHandler()

    ob.PublicEvent += handler2.handler
    ob.PublicEvent += handler2.handler

    handlers = []
    for _ in range(30):
        method = handler.handler
        ob.PublicEvent += method
        handlers.append(method)

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 300
    assert handler2.value == 20
    handler.value = 0
    handler2.value = 0

    for i in range(30):
        item = random.choice(handlers)
        handlers.remove(item)
        ob.PublicEvent -= item
        handler.value = 0
        ob.OnPublicEvent(EventArgsTest(10))
        assert handler.value == (len(handlers) * 10)
        assert handler2.value == ((i + 1) * 20)

    handler2.value = 0
    ob.OnPublicEvent(EventArgsTest(10))
    assert handler2.value == 20

    ob.PublicEvent -= handler2.handler

    handler2.value = 0
    ob.OnPublicEvent(EventArgsTest(10))
    assert handler2.value == 10

    ob.PublicEvent -= handler2.handler

    handler2.value = 0
    ob.OnPublicEvent(EventArgsTest(10))
    assert handler2.value == 0


def test_remove_internal_call_handler():
    """Test remove on an event sink implemented w/internalcall."""
    ob = EventTest()

    def h(sender, args):
        pass

    ob.PublicEvent += h
    ob.PublicEvent -= h


def test_remove_unknown_handler():
    """Test removing an event handler that was never added."""

    with pytest.raises(ValueError):
        ob = EventTest()
        handler = GenericHandler()

        ob.PublicEvent -= handler.handler


def test_handler_callback_failure():
    """Test failure mode for inappropriate handlers."""

    class BadHandler(object):
        def handler(self, one):
            return 'too many'

    ob = EventTest()
    handler = BadHandler()

    with pytest.raises(TypeError):
        ob.PublicEvent += handler.handler
        ob.OnPublicEvent(EventArgsTest(10))

    ob.PublicEvent -= handler.handler

    class BadHandler(object):
        def handler(self, one, two, three, four, five):
            return 'not enough'

    ob = EventTest()
    handler = BadHandler()

    with pytest.raises(TypeError):
        ob.PublicEvent += handler.handler
        ob.OnPublicEvent(EventArgsTest(10))

    ob.PublicEvent -= handler.handler


def test_incorrect_invokation():
    """Test incorrect invocation of events."""
    ob = EventTest()

    handler = GenericHandler()
    ob.PublicEvent += handler.handler

    with pytest.raises(TypeError):
        ob.OnPublicEvent()

    with pytest.raises(TypeError):
        ob.OnPublicEvent(32)

    ob.PublicEvent -= handler.handler


def test_explicit_cls_event_registration():
    """Test explicit CLS event registration."""
    from Python.Test import EventHandlerTest

    ob = EventTest()
    handler = GenericHandler()

    delegate = EventHandlerTest(handler.handler)
    ob.add_PublicEvent(delegate)
    assert handler.value is None

    ob.OnPublicEvent(EventArgsTest(10))
    assert handler.value == 10

    ob.remove_PublicEvent(delegate)
    assert handler.value == 10

    ob.OnPublicEvent(EventArgsTest(20))
    assert handler.value == 10


def test_implicit_cls_event_registration():
    """Test implicit CLS event registration."""

    with pytest.raises(TypeError):
        ob = EventTest()
        handler = GenericHandler()
        ob.add_PublicEvent(handler.handler)


def test_event_descriptor_abuse():
    """Test event descriptor abuse."""

    with pytest.raises(TypeError):
        del EventTest.PublicEvent

    with pytest.raises(TypeError):
        del EventTest.__dict__['PublicEvent']

    desc = EventTest.__dict__['PublicEvent']

    with pytest.raises(TypeError):
        desc.__get__(0, 0)

    with pytest.raises(TypeError):
        desc.__set__(0, 0)

    with pytest.raises(TypeError):
        ob = EventTest()
        ob.PublicEvent = 0

    with pytest.raises(TypeError):
        EventTest.PublicStaticEvent = 0



# -*- coding: utf-8 -*-

"""Test clr enum support."""

import pytest
import Python.Test as Test

from .utils import DictProxyType


def test_enum_standard_attrs():
    """Test standard enum attributes."""
    from System import DayOfWeek

    assert DayOfWeek.__name__ == 'DayOfWeek'
    assert DayOfWeek.__module__ == 'System'
    assert isinstance(DayOfWeek.__dict__, DictProxyType)
    assert DayOfWeek.__doc__ is None


def test_enum_get_member():
    """Test access to enum members."""
    from System import DayOfWeek

    assert DayOfWeek.Sunday == DayOfWeek(0)
    assert DayOfWeek.Monday == DayOfWeek(1)
    assert DayOfWeek.Tuesday == DayOfWeek(2)
    assert DayOfWeek.Wednesday == DayOfWeek(3)
    assert DayOfWeek.Thursday == DayOfWeek(4)
    assert DayOfWeek.Friday == DayOfWeek(5)
    assert DayOfWeek.Saturday == DayOfWeek(6)


def test_byte_enum():
    """Test byte enum."""
    assert Test.ByteEnum.Zero == Test.ByteEnum(0)
    assert Test.ByteEnum.One == Test.ByteEnum(1)
    assert Test.ByteEnum.Two == Test.ByteEnum(2)


def test_sbyte_enum():
    """Test sbyte enum."""
    assert Test.SByteEnum.Zero == Test.SByteEnum(0)
    assert Test.SByteEnum.One == Test.SByteEnum(1)
    assert Test.SByteEnum.Two == Test.SByteEnum(2)


def test_short_enum():
    """Test short enum."""
    assert Test.ShortEnum.Zero == Test.ShortEnum(0)
    assert Test.ShortEnum.One == Test.ShortEnum(1)
    assert Test.ShortEnum.Two == Test.ShortEnum(2)


def test_ushort_enum():
    """Test ushort enum."""
    assert Test.UShortEnum.Zero == Test.UShortEnum(0)
    assert Test.UShortEnum.One == Test.UShortEnum(1)
    assert Test.UShortEnum.Two == Test.UShortEnum(2)


def test_int_enum():
    """Test int enum."""
    assert Test.IntEnum.Zero == Test.IntEnum(0)
    assert Test.IntEnum.One == Test.IntEnum(1)
    assert Test.IntEnum.Two == Test.IntEnum(2)


def test_uint_enum():
    """Test uint enum."""
    assert Test.UIntEnum.Zero == Test.UIntEnum(0)
    assert Test.UIntEnum.One == Test.UIntEnum(1)
    assert Test.UIntEnum.Two == Test.UIntEnum(2)


def test_long_enum():
    """Test long enum."""
    assert Test.LongEnum.Zero == Test.LongEnum(0)
    assert Test.LongEnum.One == Test.LongEnum(1)
    assert Test.LongEnum.Two == Test.LongEnum(2)


def test_ulong_enum():
    """Test ulong enum."""
    assert Test.ULongEnum.Zero == Test.ULongEnum(0)
    assert Test.ULongEnum.One == Test.ULongEnum(1)
    assert Test.ULongEnum.Two == Test.ULongEnum(2)


def test_instantiate_enum_fails():
    """Test that instantiation of an enum class fails."""
    from System import DayOfWeek

    with pytest.raises(TypeError):
        _ = DayOfWeek()


def test_subclass_enum_fails():
    """Test that subclassing of an enumeration fails."""
    from System import DayOfWeek

    with pytest.raises(TypeError):
        class Boom(DayOfWeek):
            pass

        _ = Boom


def test_enum_set_member_fails():
    """Test that setattr operations on enumerations fail."""
    from System import DayOfWeek

    with pytest.raises(TypeError):
        DayOfWeek.Sunday = 13

    with pytest.raises(TypeError):
        del DayOfWeek.Sunday


def test_enum_undefined_value():
    """Test enumeration conversion with FlagsAttribute set."""
    # This works because the FlagsField enum has FlagsAttribute.
    Test.FieldTest().FlagsField = Test.FlagsEnum(99)

    # This should fail because our test enum doesn't have it.
    with pytest.raises(ValueError):
        Test.FieldTest().EnumField = Test.ShortEnum(20)
    
    # explicitly permit undefined values
    Test.FieldTest().EnumField = Test.ShortEnum(20, True)

def test_enum_conversion():
    """Test enumeration conversion."""
    ob = Test.FieldTest()
    assert ob.EnumField == Test.ShortEnum(0)

    ob.EnumField = Test.ShortEnum.One
    assert ob.EnumField == Test.ShortEnum(1)

    with pytest.raises(OverflowError):
        Test.FieldTest().EnumField = Test.ShortEnum(100000)

    with pytest.raises(TypeError):
        Test.FieldTest().EnumField = "str"
    
    with pytest.raises(TypeError):
        Test.FieldTest().EnumField = 1



# -*- coding: utf-8 -*-

"""Test PythonEngine embedding APIs."""

import sys

import System
import pytest
from Python.Runtime import PythonEngine


def test_multiple_calls_to_initialize():
    """Test that multiple initialize calls are harmless."""
    try:
        PythonEngine.Initialize()
        PythonEngine.Initialize()
        PythonEngine.Initialize()
    except Exception:
        assert False  # Initialize() raise an exception.


@pytest.mark.skip(reason="FIXME: test crashes")
def test_import_module():
    """Test module import."""
    m = PythonEngine.ImportModule("sys")
    n = m.GetAttr("__name__")
    assert n.AsManagedObject(System.String) == "sys"


@pytest.mark.skip(reason="FIXME: test freezes")
def test_run_string():
    """Test the RunString method."""
    PythonEngine.AcquireLock()

    code = "import sys; sys.singleline_worked = 1"
    PythonEngine.RunString(code)
    assert sys.singleline_worked == 1

    code = "import sys\nsys.multiline_worked = 1"
    PythonEngine.RunString(code)
    assert sys.multiline_worked == 1

    PythonEngine.ReleaseLock()



# -*- coding: utf-8 -*-

"""Test CLR method support."""

import System
import pytest
from Python.Test import MethodTest

def test_instance_method_overwritable():
    """Test instance method overwriting."""

    ob = MethodTest()
    ob.OverwritableMethod = lambda: "overwritten"
    assert ob.OverwritableMethod() == "overwritten"


def test_public_instance_method():
    """Test public instance method visibility."""
    ob = MethodTest()
    assert ob.PublicMethod() == "public"


def test_public_static_method():
    """Test public static method visibility."""
    ob = MethodTest()
    assert MethodTest.PublicStaticMethod() == "public static"
    assert ob.PublicStaticMethod() == "public static"


def test_protected_instance_method():
    """Test protected instance method visibility."""
    ob = MethodTest()
    assert ob.ProtectedMethod() == "protected"


def test_protected_static_method():
    """Test protected static method visibility."""
    ob = MethodTest()
    result = "protected static"
    assert MethodTest.ProtectedStaticMethod() == result
    assert ob.ProtectedStaticMethod() == result


def test_internal_method():
    """Test internal method visibility."""

    with pytest.raises(AttributeError):
        _ = MethodTest().InternalMethod

    with pytest.raises(AttributeError):
        _ = MethodTest.InternalMethod

    with pytest.raises(AttributeError):
        _ = MethodTest().InternalStaticMethod

    with pytest.raises(AttributeError):
        _ = MethodTest.InternalStaticMethod


def test_private_method():
    """Test private method visibility."""

    with pytest.raises(AttributeError):
        _ = MethodTest().PrivateMethod

    with pytest.raises(AttributeError):
        _ = MethodTest.PrivateMethod

    with pytest.raises(AttributeError):
        _ = MethodTest().PrivateStaticMethod

    with pytest.raises(AttributeError):
        _ = MethodTest.PrivateStaticMethod


def test_unbound_managed_method_call():
    """Test calling unbound managed methods."""
    from Python.Test import MethodTestSub

    ob = MethodTest()
    assert MethodTest.PublicMethod(ob) == "public"

    with pytest.raises(TypeError):
        MethodTest.PublicMethod()

    ob = MethodTestSub()
    assert MethodTestSub.PublicMethod(ob) == "public"
    assert MethodTestSub.PublicMethod(ob, "echo") == "echo"

    with pytest.raises(TypeError):
        MethodTestSub.PublicMethod("echo")


def test_overloaded_method_inheritance():
    """Test that overloads are inherited properly."""
    from Python.Test import MethodTestSub

    ob = MethodTest()
    assert ob.PublicMethod() == "public"

    with pytest.raises(TypeError):
        ob = MethodTest()
        ob.PublicMethod("echo")

    ob = MethodTestSub()
    assert ob.PublicMethod() == "public"

    assert ob.PublicMethod("echo") == "echo"


def test_method_descriptor_abuse():
    """Test method descriptor abuse."""
    desc = MethodTest.__dict__['PublicMethod']

    with pytest.raises(TypeError):
        desc.__get__(0, 0)

    with pytest.raises(AttributeError):
        desc.__set__(0, 0)


def test_method_docstrings():
    """Test standard method docstring generation"""
    method = MethodTest.GetType
    value = 'System.Type GetType()'
    assert method.__doc__ == value


# ======================================================================
# Tests of specific argument and result conversion scenarios
# ======================================================================
def test_method_call_enum_conversion():
    """Test enum conversion in method call."""
    from System import TypeCode

    ob = MethodTest()
    r = ob.TestEnumConversion(TypeCode.Int32)
    assert r == TypeCode.Int32


def test_method_call_flags_conversion():
    """Test flags conversion in method call."""
    from System.IO import FileAccess

    ob = MethodTest()
    flags = FileAccess.Read | FileAccess.Write
    r = ob.TestFlagsConversion(flags)
    assert r == flags


def test_method_call_struct_conversion():
    """Test struct conversion in method call."""
    from System import Guid

    ob = MethodTest()
    guid = Guid.NewGuid()
    temp = guid.ToString()
    r = ob.TestStructConversion(guid)
    assert r.ToString() == temp


def test_subclass_instance_conversion():
    """Test subclass instance conversion in method call."""

    class TestSubException(System.Exception):
        pass

    ob = MethodTest()
    instance = TestSubException()
    result = ob.TestSubclassConversion(instance)
    assert isinstance(result, System.Exception)


def test_null_array_conversion():
    """Test null array conversion in method call."""
    ob = MethodTest()
    r = ob.TestNullArrayConversion(None)
    assert r is None


def test_string_params_args():
    """Test use of string params."""
    result = MethodTest.TestStringParamsArg('one', 'two', 'three')
    assert result.Length == 4
    assert len(result) == 4, result
    assert result[0] == 'one'
    assert result[1] == 'two'
    assert result[2] == 'three'
    # ensures params string[] overload takes precedence over params object[]
    assert result[3] == 'tail'

    result = MethodTest.TestStringParamsArg(['one', 'two', 'three'])
    assert len(result) == 4
    assert result[0] == 'one'
    assert result[1] == 'two'
    assert result[2] == 'three'
    assert result[3] == 'tail'


def test_object_params_args():
    """Test use of object params."""
    result = MethodTest.TestObjectParamsArg('one', 'two', 'three')
    assert len(result) == 3, result
    assert result[0] == 'one'
    assert result[1] == 'two'
    assert result[2] == 'three'

    result = MethodTest.TestObjectParamsArg(['one', 'two', 'three'])
    assert len(result) == 3, result
    assert result[0] == 'one'
    assert result[1] == 'two'
    assert result[2] == 'three'


def test_value_params_args():
    """Test use of value type params."""
    result = MethodTest.TestValueParamsArg(1, 2, 3)
    assert len(result) == 3
    assert result[0] == 1
    assert result[1] == 2
    assert result[2] == 3

    result = MethodTest.TestValueParamsArg([1, 2, 3])
    assert len(result) == 3
    assert result[0] == 1
    assert result[1] == 2
    assert result[2] == 3


def test_non_params_array_in_last_place():
    """Test overload resolution with of non-"params" array as
    last parameter."""
    result = MethodTest.TestNonParamsArrayInLastPlace(1, 2, 3)
    assert result

def test_params_methods_with_no_params():
    """Tests that passing no arguments to a params method
    passes an empty array"""
    result = MethodTest.TestValueParamsArg()
    assert len(result) == 0

    result = MethodTest.TestOneArgWithParams('Some String')
    assert len(result) == 0

    result = MethodTest.TestTwoArgWithParams('Some String', 'Some Other String')
    assert len(result) == 0

def test_params_methods_with_non_lists():
    """Tests that passing single parameters to a params
    method will convert into an array on the .NET side"""
    result = MethodTest.TestOneArgWithParams('Some String', [1, 2, 3, 4])
    assert len(result) == 4

    result = MethodTest.TestOneArgWithParams('Some String', 1, 2, 3, 4)
    assert len(result) == 4

    result = MethodTest.TestOneArgWithParams('Some String', [5])
    assert len(result) == 1

    result = MethodTest.TestOneArgWithParams('Some String', 5)
    assert len(result) == 1

def test_params_method_with_lists():
    """Tests passing multiple lists to a params object[] method"""
    result = MethodTest.TestObjectParamsArg([],[])
    assert len(result) == 2

def test_string_out_params():
    """Test use of string out-parameters."""
    result = MethodTest.TestStringOutParams("hi", "there")
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"

    result = MethodTest.TestStringOutParams("hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"


def test_string_ref_params():
    """Test use of string byref parameters."""
    result = MethodTest.TestStringRefParams("hi", "there")
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"

    result = MethodTest.TestStringRefParams("hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"


def test_value_out_params():
    """Test use of value type out-parameters."""
    result = MethodTest.TestValueOutParams("hi", 1)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == 42

    # None cannot be converted to a value type like int, long, etc.
    with pytest.raises(TypeError):
        MethodTest.TestValueOutParams("hi", None)


def test_value_ref_params():
    """Test use of value type byref parameters."""
    result = MethodTest.TestValueRefParams("hi", 1)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == 42

    # None cannot be converted to a value type like int, long, etc.
    with pytest.raises(TypeError):
        MethodTest.TestValueRefParams("hi", None)


def test_object_out_params():
    """Test use of object out-parameters."""
    result = MethodTest.TestObjectOutParams("hi", MethodTest())
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Exception)

    result = MethodTest.TestObjectOutParams("hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Exception)


def test_object_ref_params():
    """Test use of object byref parameters."""
    result = MethodTest.TestObjectRefParams("hi", MethodTest())
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Exception)

    result = MethodTest.TestObjectRefParams("hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Exception)


def test_struct_out_params():
    """Test use of struct out-parameters."""
    result = MethodTest.TestStructOutParams("hi", System.Guid.NewGuid())
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Guid)

    # None cannot be converted to a value type like a struct
    with pytest.raises(TypeError):
        MethodTest.TestValueRefParams("hi", None)


def test_struct_ref_params():
    """Test use of struct byref parameters."""
    result = MethodTest.TestStructRefParams("hi", System.Guid.NewGuid())
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert isinstance(result[1], System.Guid)

    # None cannot be converted to a value type like a struct
    with pytest.raises(TypeError):
        MethodTest.TestValueRefParams("hi", None)


def test_void_single_out_param():
    """Test void method with single out-parameter."""
    result = MethodTest.TestVoidSingleOutParam(9)
    assert result == 42

    # None cannot be converted to a value type
    with pytest.raises(TypeError):
        MethodTest.TestVoidSingleOutParam(None)


def test_void_single_ref_param():
    """Test void method with single ref-parameter."""
    result = MethodTest.TestVoidSingleRefParam(9)
    assert result == 42

    # None cannot be converted to a value type
    with pytest.raises(TypeError):
        MethodTest.TestVoidSingleRefParam(None)


def test_single_default_param():
    """Test void method with single ref-parameter."""
    result = MethodTest.TestSingleDefaultParam()
    assert result == 5


def test_one_arg_and_two_default_param():
    """Test void method with single ref-parameter."""
    result = MethodTest.TestOneArgAndTwoDefaultParam(11)
    assert result == 22

    result = MethodTest.TestOneArgAndTwoDefaultParam(15)
    assert result == 26

    result = MethodTest.TestOneArgAndTwoDefaultParam(20)
    assert result == 31


def test_two_default_param():
    """Test void method with single ref-parameter."""
    result = MethodTest.TestTwoDefaultParam()
    assert result == 11


def test_explicit_selection_with_out_modifier():
    """Check explicit overload selection with out modifiers."""
    refstr = System.String("").GetType().MakeByRefType()
    result = MethodTest.TestStringOutParams.__overloads__[str, refstr](
        "hi", "there")
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"

    result = MethodTest.TestStringOutParams.__overloads__[str, refstr](
        "hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"


def test_explicit_selection_with_ref_modifier():
    """Check explicit overload selection with ref modifiers."""
    refstr = System.String("").GetType().MakeByRefType()
    result = MethodTest.TestStringRefParams.__overloads__[str, refstr](
        "hi", "there")
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"

    result = MethodTest.TestStringRefParams.__overloads__[str, refstr](
        "hi", None)
    assert isinstance(result, tuple)
    assert len(result) == 2
    assert result[0] is True
    assert result[1] == "output string"


def test_explicit_overload_selection():
    """Check explicit overload selection using [] syntax."""
    from Python.Test import ISayHello1, InterfaceTest, ShortEnum
    from System import Array

    inst = InterfaceTest()

    value = MethodTest.Overloaded.__overloads__[System.Boolean](True)
    assert value is True

    value = MethodTest.Overloaded.__overloads__[bool](True)
    assert value is True

    value = MethodTest.Overloaded.__overloads__[System.Byte](255)
    assert value == 255

    value = MethodTest.Overloaded.__overloads__[System.SByte](127)
    assert value == 127

    value = MethodTest.Overloaded.__overloads__[System.Char](u'A')
    assert value == u'A'

    value = MethodTest.Overloaded.__overloads__[System.Char](65535)
    assert value == chr(65535)

    value = MethodTest.Overloaded.__overloads__[System.Int16](32767)
    assert value == 32767

    value = MethodTest.Overloaded.__overloads__[System.Int32](2147483647)
    assert value == 2147483647

    value = MethodTest.Overloaded.__overloads__[int](2147483647)
    assert value == 2147483647

    value = MethodTest.Overloaded.__overloads__[System.Int64](
        9223372036854775807
    )
    assert value == 9223372036854775807

    value = MethodTest.Overloaded.__overloads__[System.UInt16](65000)
    assert value == 65000

    value = MethodTest.Overloaded.__overloads__[System.UInt32](
        4294967295
    )
    assert value == 4294967295

    value = MethodTest.Overloaded.__overloads__[System.UInt64](
        18446744073709551615
    )
    assert value == 18446744073709551615

    value = MethodTest.Overloaded.__overloads__[System.Single](3.402823e38)
    assert value == System.Single(3.402823e38)

    value = MethodTest.Overloaded.__overloads__[System.Double](
        1.7976931348623157e308)
    assert value == 1.7976931348623157e308

    value = MethodTest.Overloaded.__overloads__[float](
        1.7976931348623157e308)
    assert value == 1.7976931348623157e308

    value = MethodTest.Overloaded.__overloads__[System.Decimal](
        System.Decimal.One)
    assert value == System.Decimal.One

    value = MethodTest.Overloaded.__overloads__[System.String]("spam")
    assert value == "spam"

    value = MethodTest.Overloaded.__overloads__[str]("spam")
    assert value == "spam"

    value = MethodTest.Overloaded.__overloads__[ShortEnum](ShortEnum.Zero)
    assert value == ShortEnum.Zero

    value = MethodTest.Overloaded.__overloads__[System.Object](inst)
    assert value.__class__ == inst.__class__

    value = MethodTest.Overloaded.__overloads__[InterfaceTest](inst)
    assert value.__class__ == inst.__class__

    iface_class = ISayHello1(InterfaceTest()).__class__
    value = MethodTest.Overloaded.__overloads__[ISayHello1](inst)
    assert value.__class__ != inst.__class__
    assert value.__class__ == iface_class

    atype = Array[System.Object]
    value = MethodTest.Overloaded.__overloads__[str, int, atype](
        "one", 1, atype([1, 2, 3]))
    assert value == 3

    value = MethodTest.Overloaded.__overloads__[str, int]("one", 1)
    assert value == 1

    value = MethodTest.Overloaded.__overloads__[int, str](1, "one")
    assert value == 1


def test_overload_selection_with_array_types():
    """Check overload selection using array types."""
    from Python.Test import ISayHello1, InterfaceTest, ShortEnum
    from System import Array

    inst = InterfaceTest()

    vtype = Array[System.Boolean]
    input_ = vtype([True, True])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] is True
    assert value[1] is True

    vtype = Array[bool]
    input_ = vtype([True, True])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] is True
    assert value[1] is True

    vtype = Array[System.Byte]
    input_ = vtype([0, 255])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 255

    vtype = Array[System.SByte]
    input_ = vtype([0, 127])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 127

    vtype = Array[System.Char]
    input_ = vtype([u'A', u'Z'])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == u'A'
    assert value[1] == u'Z'

    vtype = Array[System.Char]
    input_ = vtype([0, 65535])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == chr(0)
    assert value[1] == chr(65535)

    vtype = Array[System.Int16]
    input_ = vtype([0, 32767])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 32767

    vtype = Array[System.Int32]
    input_ = vtype([0, 2147483647])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 2147483647

    vtype = Array[int]
    input_ = vtype([0, 2147483647])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 2147483647

    vtype = Array[System.Int64]
    input_ = vtype([0, 9223372036854775807])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 9223372036854775807

    vtype = Array[System.UInt16]
    input_ = vtype([0, 65000])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 65000

    vtype = Array[System.UInt32]
    input_ = vtype([0, 4294967295])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 4294967295

    vtype = Array[System.UInt64]
    input_ = vtype([0, 18446744073709551615])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0
    assert value[1] == 18446744073709551615

    vtype = Array[System.Single]
    input_ = vtype([0.0, 3.402823e38])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0.0
    assert value[1] == System.Single(3.402823e38)

    vtype = Array[System.Double]
    input_ = vtype([0.0, 1.7976931348623157e308])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0.0
    assert value[1] == 1.7976931348623157e308

    vtype = Array[float]
    input_ = vtype([0.0, 1.7976931348623157e308])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == 0.0
    assert value[1] == 1.7976931348623157e308

    vtype = Array[System.Decimal]
    input_ = vtype([System.Decimal.Zero, System.Decimal.One])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == System.Decimal.Zero
    assert value[1] == System.Decimal.One

    vtype = Array[System.String]
    input_ = vtype(["one", "two"])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == "one"
    assert value[1] == "two"

    vtype = Array[str]
    input_ = vtype(["one", "two"])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == "one"
    assert value[1] == "two"

    vtype = Array[ShortEnum]
    input_ = vtype([ShortEnum.Zero, ShortEnum.One])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0] == ShortEnum.Zero
    assert value[1] == ShortEnum.One

    vtype = Array[System.Object]
    input_ = vtype([inst, inst])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].__class__ == inst.__class__
    assert value[1].__class__ == inst.__class__

    vtype = Array[InterfaceTest]
    input_ = vtype([inst, inst])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].__class__ == inst.__class__
    assert value[1].__class__ == inst.__class__

    iface_class = ISayHello1(inst).__class__
    vtype = Array[ISayHello1]
    input_ = vtype([inst, inst])
    value = MethodTest.Overloaded.__overloads__[vtype](input_)
    assert value[0].__class__ == iface_class
    assert value[1].__class__ == iface_class


def test_explicit_overload_selection_failure():
    """Check that overload selection fails correctly."""

    with pytest.raises(TypeError):
        _ = MethodTest.Overloaded.__overloads__[System.Type](True)

    with pytest.raises(TypeError):
        _ = MethodTest.Overloaded.__overloads__[int, int](1, 1)

    with pytest.raises(TypeError):
        _ = MethodTest.Overloaded.__overloads__[str, int, int]("", 1, 1)

    with pytest.raises(TypeError):
        _ = MethodTest.Overloaded.__overloads__[int, int](1)


def test_we_can_bind_to_encoding_get_string():
    """Check that we can bind to the Encoding.GetString method
    with variables."""
    from System.Text import Encoding
    from System.IO import MemoryStream

    my_bytes = Encoding.UTF8.GetBytes('Some testing string')
    stream = MemoryStream()
    stream.Write(my_bytes, 0, my_bytes.Length)
    stream.Position = 0

    buff = System.Array.CreateInstance(System.Byte, 3)
    buff.Initialize()
    data = []
    read = 1

    while read > 0:
        read = stream.Read(buff, 0, buff.Length)
        temp = Encoding.UTF8.GetString(buff, 0, read)
        data.append(temp)

    data = ''.join(data)
    assert data == 'Some testing string'


def test_wrong_overload():
    """Test regression in which implicit conversion caused the wrong types
    to be used. See #131 for issue. Fixed by #137, #151"""

    # Used to return `50L`
    res = System.Math.Abs(50.5)
    assert res == 50.5
    assert type(res) == float

    res = System.Math.Abs(-50.5)
    assert res == 50.5
    assert type(res) == float

    res = System.Math.Max(50.5, 50.1)
    assert res == 50.5
    assert type(res) == float

    res = System.Math.Max(System.Double(10.5), System.Double(50.5))
    assert res == 50.5
    assert type(res) == float  # Should it return a System.Double?

    res = System.Math.Max(System.Double(50.5), 50.1)
    assert res == 50.5
    assert type(res) == float


def test_no_object_in_param():
    """Test that fix for #203 doesn't break behavior w/ no object overload"""

    res = MethodTest.TestOverloadedNoObject(5)
    assert res == "Got int"

    res = MethodTest.TestOverloadedNoObject(i=7)
    assert res == "Got int"

    with pytest.raises(TypeError):
        MethodTest.TestOverloadedNoObject("test")

    with pytest.raises(TypeError):
        MethodTest.TestOverloadedNoObject(5.5)

    # Ensure that the top-level error is TypeError even if the inner error is an OverflowError
    with pytest.raises(TypeError):
        MethodTest.TestOverloadedNoObject(2147483648)

def test_object_in_param():
    """Test regression introduced by #151 in which Object method overloads
    aren't being used. See #203 for issue."""

    res = MethodTest.TestOverloadedObject(5)
    assert res == "Got int"

    res = MethodTest.TestOverloadedObject(i=7)
    assert res == "Got int"

    res = MethodTest.TestOverloadedObject("test")
    assert res == "Got object"

    res = MethodTest.TestOverloadedObject(o="test")
    assert res == "Got object"


def test_object_in_multiparam():
    """Test method with object multiparams behaves"""

    res = MethodTest.TestOverloadedObjectTwo(5, 5)
    assert res == "Got int-int"

    res = MethodTest.TestOverloadedObjectTwo(5, "foo")
    assert res == "Got int-string"

    res = MethodTest.TestOverloadedObjectTwo("foo", 7.24)
    assert res == "Got string-object"

    res = MethodTest.TestOverloadedObjectTwo("foo", "bar")
    assert res == "Got string-string"

    res = MethodTest.TestOverloadedObjectTwo("foo", 5)
    assert res == "Got string-int"

    res = MethodTest.TestOverloadedObjectTwo(7.24, 7.24)
    assert res == "Got object-object"

    res = MethodTest.TestOverloadedObjectTwo(a=5, b=5)
    assert res == "Got int-int"

    res = MethodTest.TestOverloadedObjectTwo(5, b=5)
    assert res == "Got int-int"

    res = MethodTest.TestOverloadedObjectTwo(a=5, b="foo")
    assert res == "Got int-string"

    res = MethodTest.TestOverloadedObjectTwo(5, b="foo")
    assert res == "Got int-string"

    res = MethodTest.TestOverloadedObjectTwo(a="foo", b=7.24)
    assert res == "Got string-object"

    res = MethodTest.TestOverloadedObjectTwo("foo", b=7.24)
    assert res == "Got string-object"

    res = MethodTest.TestOverloadedObjectTwo(a="foo", b="bar")
    assert res == "Got string-string"

    res = MethodTest.TestOverloadedObjectTwo("foo", b="bar")
    assert res == "Got string-string"

    res = MethodTest.TestOverloadedObjectTwo(a="foo", b=5)
    assert res == "Got string-int"

    res = MethodTest.TestOverloadedObjectTwo("foo", b=5)
    assert res == "Got string-int"

    res = MethodTest.TestOverloadedObjectTwo(a=7.24, b=7.24)
    assert res == "Got object-object"

    res = MethodTest.TestOverloadedObjectTwo(7.24, b=7.24)
    assert res == "Got object-object"


def test_object_in_multiparam_exception():
    """Test method with object multiparams behaves"""

    with pytest.raises(TypeError) as excinfo:
        MethodTest.TestOverloadedObjectThree("foo", "bar")

    e = excinfo.value
    c = e.__cause__
    assert c.GetType().FullName == 'System.AggregateException'
    assert len(c.InnerExceptions) == 2

def test_case_sensitive():
    """Test that case-sensitivity is respected. GH#81"""

    res = MethodTest.CaseSensitive()
    assert res == "CaseSensitive"

    res = MethodTest.Casesensitive()
    assert res == "Casesensitive"

    with pytest.raises(AttributeError):
        MethodTest.casesensitive()

def test_getting_generic_method_binding_does_not_leak_ref_count():
    """Test that managed object is freed after calling generic method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import sys

    refCount = sys.getrefcount(PlainOldClass().GenericMethod[str])
    assert refCount == 1

def test_getting_generic_method_binding_does_not_leak_memory():
    """Test that managed object is freed after calling generic method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import psutil, os, gc, clr

    process = psutil.Process(os.getpid())
    processBytesBeforeCall = process.memory_info().rss
    print("\n\nMemory consumption (bytes) at start of test: " + str(processBytesBeforeCall))

    iterations = 500
    for i in range(iterations):
        PlainOldClass().GenericMethod[str]

    gc.collect()
    System.GC.Collect()

    processBytesAfterCall = process.memory_info().rss
    print("Memory consumption (bytes) at end of test: " + str(processBytesAfterCall))
    processBytesDelta = processBytesAfterCall - processBytesBeforeCall
    print("Memory delta: " + str(processBytesDelta))

    bytesAllocatedPerIteration = pow(2, 20)  # 1MB
    bytesLeakedPerIteration = processBytesDelta / iterations

    # Allow 50% threshold - this shows the original issue is fixed, which leaks the full allocated bytes per iteration
    failThresholdBytesLeakedPerIteration = bytesAllocatedPerIteration / 2

    assert bytesLeakedPerIteration < failThresholdBytesLeakedPerIteration

def test_getting_overloaded_method_binding_does_not_leak_ref_count():
    """Test that managed object is freed after calling overloaded method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import sys

    refCount = sys.getrefcount(PlainOldClass().OverloadedMethod.Overloads[int])
    assert refCount == 1

def test_getting_overloaded_method_binding_does_not_leak_memory():
    """Test that managed object is freed after calling overloaded method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import psutil, os, gc, clr

    process = psutil.Process(os.getpid())
    processBytesBeforeCall = process.memory_info().rss
    print("\n\nMemory consumption (bytes) at start of test: " + str(processBytesBeforeCall))

    iterations = 500
    for i in range(iterations):
        PlainOldClass().OverloadedMethod.Overloads[int]

    gc.collect()
    System.GC.Collect()

    processBytesAfterCall = process.memory_info().rss
    print("Memory consumption (bytes) at end of test: " + str(processBytesAfterCall))
    processBytesDelta = processBytesAfterCall - processBytesBeforeCall
    print("Memory delta: " + str(processBytesDelta))

    bytesAllocatedPerIteration = pow(2, 20)  # 1MB
    bytesLeakedPerIteration = processBytesDelta / iterations

    # Allow 50% threshold - this shows the original issue is fixed, which leaks the full allocated bytes per iteration
    failThresholdBytesLeakedPerIteration = bytesAllocatedPerIteration / 2

    assert bytesLeakedPerIteration < failThresholdBytesLeakedPerIteration

def test_getting_method_overloads_binding_does_not_leak_ref_count():
    """Test that managed object is freed after calling overloaded method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import sys

    refCount = sys.getrefcount(PlainOldClass().OverloadedMethod.Overloads)
    assert refCount == 1

def test_getting_method_overloads_binding_does_not_leak_memory():
    """Test that managed object is freed after calling overloaded method. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import psutil, os, gc, clr

    process = psutil.Process(os.getpid())
    processBytesBeforeCall = process.memory_info().rss
    print("\n\nMemory consumption (bytes) at start of test: " + str(processBytesBeforeCall))

    iterations = 500
    for i in range(iterations):
        PlainOldClass().OverloadedMethod.Overloads

    gc.collect()
    System.GC.Collect()

    processBytesAfterCall = process.memory_info().rss
    print("Memory consumption (bytes) at end of test: " + str(processBytesAfterCall))
    processBytesDelta = processBytesAfterCall - processBytesBeforeCall
    print("Memory delta: " + str(processBytesDelta))

    bytesAllocatedPerIteration = pow(2, 20)  # 1MB
    bytesLeakedPerIteration = processBytesDelta / iterations

    # Allow 50% threshold - this shows the original issue is fixed, which leaks the full allocated bytes per iteration
    failThresholdBytesLeakedPerIteration = bytesAllocatedPerIteration / 2

    assert bytesLeakedPerIteration < failThresholdBytesLeakedPerIteration

def test_getting_overloaded_constructor_binding_does_not_leak_ref_count():
    """Test that managed object is freed after calling overloaded constructor, constructorbinding.cs mp_subscript. Issue #691"""

    from PlainOldNamespace import PlainOldClass

    import sys

    # simple test
    refCount = sys.getrefcount(PlainOldClass.Overloads[int])
    assert refCount == 1


def test_default_params():
    # all positional parameters
    res = MethodTest.DefaultParams(1,2,3,4)
    assert res == "1234"

    res = MethodTest.DefaultParams(1, 2, 3)
    assert res == "1230"

    res = MethodTest.DefaultParams(1, 2)
    assert res == "1200"

    res = MethodTest.DefaultParams(1)
    assert res == "1000"

    res = MethodTest.DefaultParams(a=2)
    assert res == "2000"

    res = MethodTest.DefaultParams(b=3)
    assert res == "0300"

    res = MethodTest.DefaultParams(c=4)
    assert res == "0040"

    res = MethodTest.DefaultParams(d=7)
    assert res == "0007"

    res = MethodTest.DefaultParams(a=2, c=5)
    assert res == "2050"

    res = MethodTest.DefaultParams(1, d=7, c=3)
    assert res == "1037"

    with pytest.raises(TypeError):
        MethodTest.DefaultParams(1,2,3,4,5)

def test_optional_params():
    res = MethodTest.OptionalParams(1, 2, 3, 4)
    assert res == "1234"

    res = MethodTest.OptionalParams(1, 2, 3)
    assert res == "1230"

    res = MethodTest.OptionalParams(1, 2)
    assert res == "1200"

    res = MethodTest.OptionalParams(1)
    assert res == "1000"

    res = MethodTest.OptionalParams(a=2)
    assert res == "2000"

    res = MethodTest.OptionalParams(b=3)
    assert res == "0300"

    res = MethodTest.OptionalParams(c=4)
    assert res == "0040"

    res = MethodTest.OptionalParams(d=7)
    assert res == "0007"

    res = MethodTest.OptionalParams(a=2, c=5)
    assert res == "2050"

    res = MethodTest.OptionalParams(1, d=7, c=3)
    assert res == "1037"

    res = MethodTest.OptionalParams_TestMissing()
    assert res == True

    res = MethodTest.OptionalParams_TestMissing(None)
    assert res == False

    res = MethodTest.OptionalParams_TestMissing(a = None)
    assert res == False

    res = MethodTest.OptionalParams_TestMissing(a='hi')
    assert res == False

    res = MethodTest.OptionalParams_TestReferenceType()
    assert res == True

    res = MethodTest.OptionalParams_TestReferenceType(None)
    assert res == True

    res = MethodTest.OptionalParams_TestReferenceType(a=None)
    assert res == True

    res = MethodTest.OptionalParams_TestReferenceType('hi')
    assert res == False

    res = MethodTest.OptionalParams_TestReferenceType(a='hi')
    assert res == False

def test_optional_and_default_params():

    res = MethodTest.OptionalAndDefaultParams()
    assert res == "0000"

    res = MethodTest.OptionalAndDefaultParams(1)
    assert res == "1000"

    res = MethodTest.OptionalAndDefaultParams(1, c=4)
    assert res == "1040"

    res = MethodTest.OptionalAndDefaultParams(b=4, c=7)
    assert res == "0470"

    res = MethodTest.OptionalAndDefaultParams2()
    assert res == "0012"

    res = MethodTest.OptionalAndDefaultParams2(a=1,b=2,c=3,d=4)
    assert res == "1234"

    res = MethodTest.OptionalAndDefaultParams2(b=2, c=3)
    assert res == "0232"

def test_default_params_overloads():
    res = MethodTest.DefaultParamsWithOverloading(1, 2)
    assert res == "12"

    res = MethodTest.DefaultParamsWithOverloading(b=5)
    assert res == "25"

    res = MethodTest.DefaultParamsWithOverloading("d")
    assert res == "dbX"

    res = MethodTest.DefaultParamsWithOverloading(b="c")
    assert res == "acX"

    res = MethodTest.DefaultParamsWithOverloading(c=3)
    assert res == "013XX"

    res = MethodTest.DefaultParamsWithOverloading(5, c=2)
    assert res == "512XX"

    res = MethodTest.DefaultParamsWithOverloading(c=0, d=1)
    assert res == "5601XXX"

    res = MethodTest.DefaultParamsWithOverloading(1, d=1)
    assert res == "1671XXX"

def test_default_params_overloads_ambiguous_call():
    with pytest.raises(TypeError):
        MethodTest.DefaultParamsWithOverloading()

def test_keyword_arg_method_resolution():
    from Python.Test import MethodArityTest

    ob = MethodArityTest()
    assert ob.Foo(1, b=2) == "Arity 2"

def test_params_array_overload():
    res = MethodTest.ParamsArrayOverloaded()
    assert res == "without params-array"

    res = MethodTest.ParamsArrayOverloaded(1)
    assert res == "without params-array"

    res = MethodTest.ParamsArrayOverloaded(i=1)
    assert res == "without params-array"

    res = MethodTest.ParamsArrayOverloaded(1, 2)
    assert res == "with params-array"

    res = MethodTest.ParamsArrayOverloaded(1, 2, 3)
    assert res == "with params-array"

    res = MethodTest.ParamsArrayOverloaded(1, paramsArray=[])
    assert res == "with params-array"

    res = MethodTest.ParamsArrayOverloaded(1, i=1)
    assert res == "with params-array"

    res = MethodTest.ParamsArrayOverloaded(1, 2, 3, i=1)
    assert res == "with params-array"

@pytest.mark.skip(reason="FIXME: incorrectly failing")
def test_params_array_overloaded_failing():
    res = MethodTest.ParamsArrayOverloaded(1, 2, i=1)
    assert res == "with params-array"

    res = MethodTest.ParamsArrayOverloaded(paramsArray=[], i=1)
    assert res == "with params-array"

def test_method_encoding():
    MethodTest.EncodingTestÅngström()

def test_method_with_pointer_array_argument():
    with pytest.raises(TypeError):
        MethodTest.PointerArray([0])



# -*- coding: utf-8 -*-

"""Test CLR property support."""

import pytest
from Python.Test import PropertyTest


def test_public_instance_property():
    """Test public instance properties."""
    ob = PropertyTest()

    assert ob.PublicProperty == 0
    ob.PublicProperty = 1
    assert ob.PublicProperty == 1

    with pytest.raises(TypeError):
        del PropertyTest().PublicProperty


def test_public_static_property():
    """Test public static properties."""
    ob = PropertyTest()

    assert PropertyTest.PublicStaticProperty == 0
    PropertyTest.PublicStaticProperty = 1
    assert PropertyTest.PublicStaticProperty == 1

    assert ob.PublicStaticProperty == 1
    ob.PublicStaticProperty = 0
    assert ob.PublicStaticProperty == 0

    with pytest.raises(TypeError):
        del PropertyTest.PublicStaticProperty

    with pytest.raises(TypeError):
        del PropertyTest().PublicStaticProperty


def test_protected_instance_property():
    """Test protected instance properties."""
    ob = PropertyTest()

    assert ob.ProtectedProperty == 0
    ob.ProtectedProperty = 1
    assert ob.ProtectedProperty == 1

    with pytest.raises(TypeError):
        del PropertyTest().ProtectedProperty


def test_protected_static_property():
    """Test protected static properties."""
    ob = PropertyTest()

    assert PropertyTest.ProtectedStaticProperty == 0
    PropertyTest.ProtectedStaticProperty = 1
    assert PropertyTest.ProtectedStaticProperty == 1

    assert ob.ProtectedStaticProperty == 1
    ob.ProtectedStaticProperty = 0
    assert ob.ProtectedStaticProperty == 0

    with pytest.raises(TypeError):
        del PropertyTest.ProtectedStaticProperty

    with pytest.raises(TypeError):
        del PropertyTest().ProtectedStaticProperty


def test_internal_property():
    """Test internal properties."""

    with pytest.raises(AttributeError):
        _ = PropertyTest().InternalProperty

    with pytest.raises(AttributeError):
        _ = PropertyTest().InternalStaticProperty

    with pytest.raises(AttributeError):
        _ = PropertyTest.InternalStaticProperty


def test_private_property():
    """Test private properties."""

    with pytest.raises(AttributeError):
        _ = PropertyTest().PrivateProperty

    with pytest.raises(AttributeError):
        _ = PropertyTest().PrivateStaticProperty

    with pytest.raises(AttributeError):
        _ = PropertyTest.PrivateStaticProperty


def test_property_descriptor_get_set():
    """Test property descriptor get / set."""

    # This test ensures that setting an attribute implemented with
    # a descriptor actually goes through the descriptor (rather than
    # silently replacing the descriptor in the instance or type dict.

    ob = PropertyTest()

    assert PropertyTest.PublicStaticProperty == 0
    assert ob.PublicStaticProperty == 0

    descriptor = PropertyTest.__dict__['PublicStaticProperty']
    assert type(descriptor) != int

    ob.PublicStaticProperty = 0
    descriptor = PropertyTest.__dict__['PublicStaticProperty']
    assert type(descriptor) != int

    PropertyTest.PublicStaticProperty = 0
    descriptor = PropertyTest.__dict__['PublicStaticProperty']
    assert type(descriptor) != int


def test_property_descriptor_wrong_type():
    """Test setting a property using a value of the wrong type."""

    with pytest.raises(TypeError):
        ob = PropertyTest()
        ob.PublicProperty = "spam"


def test_property_descriptor_abuse():
    """Test property descriptor abuse."""
    desc = PropertyTest.__dict__['PublicProperty']

    with pytest.raises(TypeError):
        desc.__get__(0, 0)

    with pytest.raises(TypeError):
        desc.__set__(0, 0)


def test_interface_property():
    """Test properties of interfaces. Added after a bug report
       that an IsAbstract check was inappropriate and prevented
       use of properties when only the interface is known."""
    from System.Collections import Hashtable, ICollection

    mapping = Hashtable()
    coll = ICollection(mapping)
    assert coll.Count == 0



#!/usr/bin/env python

from setuptools import setup, Command, Extension
from setuptools.command.build_ext import build_ext
import distutils
from distutils.command import build
from subprocess import check_output, check_call

import sys, os

PY_MAJOR = sys.version_info[0]
PY_MINOR = sys.version_info[1]

# Disable SourceLink during the build until it can read repo-format v1, #1613
os.environ["EnableSourceControlManagerQueries"] = "false"

class DotnetLib:
    def __init__(self, name, path, **kwargs):
        self.name = name
        self.path = path
        self.args = kwargs


class build_dotnet(Command):
    """Build command for dotnet-cli based builds"""

    description = "Build DLLs with dotnet-cli"
    user_options = [
        ("dotnet-config=", None, "dotnet build configuration"),
        (
            "inplace",
            "i",
            "ignore build-lib and put compiled extensions into the source "
            + "directory alongside your pure Python modules",
        ),
    ]

    def initialize_options(self):
        self.dotnet_config = None
        self.build_lib = None
        self.inplace = False

    def finalize_options(self):
        if self.dotnet_config is None:
            self.dotnet_config = "release"

        build = self.distribution.get_command_obj("build")
        build.ensure_finalized()
        if self.inplace:
            self.build_lib = "."
        else:
            self.build_lib = build.build_lib

    def run(self):
        dotnet_modules = self.distribution.dotnet_libs

        for lib in dotnet_modules:
            output = os.path.join(
                os.path.abspath(self.build_lib), lib.args.pop("output")
            )
            rename = lib.args.pop("rename", {})

            opts = sum(
                [
                    ["--" + name.replace("_", "-"), value]
                    for name, value in lib.args.items()
                ],
                [],
            )

            opts.extend(["--configuration", self.dotnet_config])
            opts.extend(["--output", output])

            self.announce("Running dotnet build...", level=distutils.log.INFO)
            self.spawn(["dotnet", "build", lib.path] + opts)

            for k, v in rename.items():
                source = os.path.join(output, k)
                dest = os.path.join(output, v)

                if os.path.isfile(source):
                    try:
                        os.remove(dest)
                    except OSError:
                        pass

                    self.move_file(src=source, dst=dest, level=distutils.log.INFO)
                else:
                    self.warn(
                        "Can't find file to rename: {}, current dir: {}".format(
                            source, os.getcwd()
                        )
                    )


# Add build_dotnet to the build tasks:
from distutils.command.build import build as _build
from setuptools.command.develop import develop as _develop
from wheel.bdist_wheel import bdist_wheel as _bdist_wheel
from setuptools import Distribution
import setuptools


class build(_build):
    sub_commands = _build.sub_commands + [("build_dotnet", None)]


class develop(_develop):
    def install_for_development(self):
        # Build extensions in-place
        self.reinitialize_command("build_dotnet", inplace=1)
        self.run_command("build_dotnet")

        return super().install_for_development()


class bdist_wheel(_bdist_wheel):
    def finalize_options(self):
        # Monkey patch bdist_wheel to think the package is pure even though we
        # include DLLs
        super().finalize_options()
        self.root_is_pure = True


# Monkey-patch Distribution s.t. it supports the dotnet_libs attribute
Distribution.dotnet_libs = None

cmdclass = {
    "build": build,
    "build_dotnet": build_dotnet,
    "develop": develop,
    "bdist_wheel": bdist_wheel,
}


with open("README.rst", "r") as f:
    long_description = f.read()

dotnet_libs = [
    DotnetLib(
        "python-runtime",
        "src/runtime/Python.Runtime.csproj",
        output="pythonnet/runtime",
    )
]

setup(
    cmdclass=cmdclass,
    name="pythonnet",
    version="3.0.0-alpha2",
    description=".Net and Mono integration for Python",
    url="https://pythonnet.github.io/",
    license="MIT",
    author="The Contributors of the Python.NET Project",
    author_email="pythonnet@python.org",
    packages=["pythonnet", "pythonnet.find_libpython"],
    install_requires=["clr_loader"],
    long_description=long_description,
    long_description_content_type="text/x-rst",
    py_modules=["clr"],
    dotnet_libs=dotnet_libs,
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: C#",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Operating System :: Microsoft :: Windows",
        "Operating System :: POSIX :: Linux",
        "Operating System :: MacOS :: MacOS X",
    ],
    zip_safe=False,
)



# -*- coding: utf-8 -*-

"""Test conversions using codecs from client python code"""
import clr
import System
import pytest
import Python.Runtime
from Python.Test import ListConversionTester, ListMember

class int_iterable():
    def __init__(self):
        self.counter = 0
    def __iter__(self):
        return self
    def __next__(self):
        if self.counter == 3:
            raise StopIteration
        self.counter = self.counter + 1
        return self.counter

class obj_iterable():
    def __init__(self):
        self.counter = 0
    def __iter__(self):
        return self
    def __next__(self):
        if self.counter == 3:
            raise StopIteration
        self.counter = self.counter + 1
        return ListMember(self.counter, "Number " + str(self.counter))

def test_iterable():
    """Test that a python iterable can be passed into a function that takes an IEnumerable<object>"""

    #Python.Runtime.Codecs.ListDecoder.Register()
    #Python.Runtime.Codecs.SequenceDecoder.Register()
    Python.Runtime.Codecs.IterableDecoder.Register()
    ob = ListConversionTester()

    iterable = int_iterable()
    assert 3 == ob.GetLength(iterable)

    iterable2 = obj_iterable()
    assert 3 == ob.GetLength2(iterable2)

    Python.Runtime.PyObjectConversions.Reset()

def test_sequence():
    Python.Runtime.Codecs.SequenceDecoder.Register()
    ob = ListConversionTester()

    tup = (1,2,3)
    assert 3 == ob.GetLength(tup)

    tup2 = (ListMember(1, "one"), ListMember(2, "two"), ListMember(3, "three"))
    assert 3 == ob.GetLength(tup2)

    Python.Runtime.PyObjectConversions.Reset()

def test_list():
    Python.Runtime.Codecs.SequenceDecoder.Register()
    ob = ListConversionTester()

    l = [1,2,3]
    assert 3 == ob.GetLength(l)

    l2 = [ListMember(1, "one"), ListMember(2, "two"), ListMember(3, "three")]
    assert 3 == ob.GetLength(l2)

    Python.Runtime.PyObjectConversions.Reset()



# -*- coding: utf-8 -*-

"""Test CLR bridge threading and GIL handling."""

import threading
import time

import _thread as thread
from .utils import dprint


def test_simple_callback_to_python():
    """Test a call to managed code that then calls back into Python."""
    from Python.Test import ThreadTest

    dprint("thread %s SimpleCallBack" % thread.get_ident())
    result = ThreadTest.CallEchoString("spam")
    assert result == "spam"
    dprint("thread %s SimpleCallBack ret" % thread.get_ident())


def test_double_callback_to_python():
    """Test a call to managed code that then calls back into Python
       that then calls managed code that then calls Python again."""
    from Python.Test import ThreadTest

    dprint("thread %s DoubleCallBack" % thread.get_ident())
    result = ThreadTest.CallEchoString2("spam")
    assert result == "spam"
    dprint("thread %s DoubleCallBack ret" % thread.get_ident())


def test_python_thread_calls_to_clr():
    """Test calls by Python-spawned threads into managed code."""
    # This test is very likely to hang if something is wrong ;)
    import System

    done = []

    def run_thread():
        for i in range(10):
            time.sleep(0.1)
            dprint("thread %s %d" % (thread.get_ident(), i))
            mstr = System.String("thread %s %d" % (thread.get_ident(), i))
            dprint(mstr.ToString())
            done.append(None)
            dprint("thread %s %d done" % (thread.get_ident(), i))

    def start_threads(count):
        for _ in range(count):
            thread_ = threading.Thread(target=run_thread)
            thread_.start()

    start_threads(5)

    while len(done) < 50:
        dprint(len(done))
        time.sleep(0.1)



# -*- coding: utf-8 -*-

"""Test that callbacks from C# into python work."""


def simpleDefaultArg(arg='test'):
    return arg


def test_default_for_null():
    """Test that C# can use null for an optional python argument"""
    from Python.Test import CallbackTest

    test_instance = CallbackTest()
    ret_val = test_instance.Call_simpleDefaultArg_WithNull(__name__)
    python_ret_val = simpleDefaultArg(None)
    assert ret_val == python_ret_val


def test_default_for_none():
    """Test that C# can use no argument for an optional python argument"""
    from Python.Test import CallbackTest

    test_instance = CallbackTest()
    ret_val = test_instance.Call_simpleDefaultArg_WithEmptyArgs(__name__)
    python_ret_val = simpleDefaultArg()
    assert ret_val == python_ret_val



import System.Collections.Generic as C

def test_contains():
    l = C.List[int]()
    l.Add(42)
    assert 42 in l
    assert 43 not in l

def test_dict_items():
    d = C.Dictionary[int, str]()
    d[42] = "a"
    items = d.items()
    assert len(items) == 1
    k,v = items[0]
    assert k == 42
    assert v == "a"



# -*- coding: utf-8 -*-

"""Test support for indexer properties."""

import Python.Test as Test
import pytest


def test_public_indexer():
    """Test public indexers."""
    ob = Test.PublicIndexerTest()

    ob[0] = "zero"
    assert ob[0] == "zero"

    ob[1] = "one"
    assert ob[1] == "one"

    assert ob[10] is None


def test_protected_indexer():
    """Test protected indexers."""
    ob = Test.ProtectedIndexerTest()

    ob[0] = "zero"
    assert ob[0] == "zero"

    ob[1] = "one"
    assert ob[1] == "one"

    assert ob[10] is None


def test_internal_indexer():
    """Test internal indexers."""
    ob = Test.InternalIndexerTest()

    with pytest.raises(TypeError):
        ob[0] = "zero"

    with pytest.raises(TypeError):
        Test.InternalIndexerTest.__getitem__(ob, 0)

    with pytest.raises(AttributeError):
        ob.__getitem__(0)


def test_private_indexer():
    """Test private indexers."""
    ob = Test.PrivateIndexerTest()

    with pytest.raises(TypeError):
        ob[0] = "zero"

    with pytest.raises(TypeError):
        Test.PrivateIndexerTest.__getitem__(ob, 0)

    with pytest.raises(AttributeError):
        ob.__getitem__(0)


def test_boolean_indexer():
    """Test boolean indexers."""
    ob = Test.BooleanIndexerTest()

    assert ob[True] is None

    with pytest.raises(TypeError):
        ob[1]
    with pytest.raises(TypeError):
        ob[0]
    with pytest.raises(TypeError):
        ob[1] = "true"
    with pytest.raises(TypeError):
        ob[0] = "false"

    ob[False] = "false"
    assert ob[False] == "false"

    ob[True] = "true"
    assert ob[True] == "true"


def test_byte_indexer():
    """Test byte indexers."""
    ob = Test.ByteIndexerTest()
    max_ = 255
    min_ = 0

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.ByteIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.ByteIndexerTest()
        ob["wrong"] = "wrong"


def test_sbyte_indexer():
    """Test sbyte indexers."""
    ob = Test.SByteIndexerTest()
    max_ = 127
    min_ = -128

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.SByteIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.SByteIndexerTest()
        ob["wrong"] = "wrong"


def test_char_indexer():
    """Test char indexers."""
    ob = Test.CharIndexerTest()
    max_ = chr(65535)
    min_ = chr(0)

    assert ob[max_] is None

    ob[max_] = "max_"
    assert ob[max_] == "max_"

    ob[min_] = "min_"
    assert ob[min_] == "min_"

    with pytest.raises(TypeError):
        ob = Test.CharIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.CharIndexerTest()
        ob["wrong"] = "wrong"


def test_int16_indexer():
    """Test Int16 indexers."""
    ob = Test.Int16IndexerTest()
    max_ = 32767
    min_ = -32768

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.Int16IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int16IndexerTest()
        ob["wrong"] = "wrong"


def test_int32_indexer():
    """Test Int32 indexers."""
    ob = Test.Int32IndexerTest()
    max_ = 2147483647
    min_ = -2147483648

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.Int32IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int32IndexerTest()
        ob["wrong"] = "wrong"


def test_int64_indexer():
    """Test Int64 indexers."""
    ob = Test.Int64IndexerTest()
    max_ = 9223372036854775807
    min_ = -9223372036854775808

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.Int64IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.Int64IndexerTest()
        ob["wrong"] = "wrong"


def test_uint16_indexer():
    """Test UInt16 indexers."""
    ob = Test.UInt16IndexerTest()
    max_ = 65535
    min_ = 0

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.UInt16IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt16IndexerTest()
        ob["wrong"] = "wrong"


def test_uint32_indexer():
    """Test UInt32 indexers."""
    ob = Test.UInt32IndexerTest()
    max_ = 4294967295
    min_ = 0

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.UInt32IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt32IndexerTest()
        ob["wrong"] = "wrong"


def test_uint64_indexer():
    """Test UInt64 indexers."""
    ob = Test.UInt64IndexerTest()
    max_ = 18446744073709551615
    min_ = 0

    assert ob[max_] is None

    ob[max_] = str(max_)
    assert ob[max_] == str(max_)

    ob[min_] = str(min_)
    assert ob[min_] == str(min_)

    with pytest.raises(TypeError):
        ob = Test.UInt64IndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.UInt64IndexerTest()
        ob["wrong"] = "wrong"


def test_single_indexer():
    """Test Single indexers."""
    ob = Test.SingleIndexerTest()
    max_ = 3.402823e38
    min_ = -3.402823e38

    assert ob[max_] is None

    ob[max_] = "max_"
    assert ob[max_] == "max_"

    ob[min_] = "min_"
    assert ob[min_] == "min_"

    with pytest.raises(TypeError):
        ob = Test.SingleIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.SingleIndexerTest()
        ob["wrong"] = "wrong"


def test_double_indexer():
    """Test Double indexers."""
    ob = Test.DoubleIndexerTest()
    max_ = 1.7976931348623157e308
    min_ = -1.7976931348623157e308

    assert ob[max_] is None

    ob[max_] = "max_"
    assert ob[max_] == "max_"

    ob[min_] = "min_"
    assert ob[min_] == "min_"

    with pytest.raises(TypeError):
        ob = Test.DoubleIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.DoubleIndexerTest()
        ob["wrong"] = "wrong"


def test_decimal_indexer():
    """Test Decimal indexers."""
    ob = Test.DecimalIndexerTest()

    from System import Decimal
    max_d = Decimal.Parse("79228162514264337593543950335")
    min_d = Decimal.Parse("-79228162514264337593543950335")

    assert ob[max_d] is None

    ob[max_d] = "max_"
    assert ob[max_d] == "max_"

    ob[min_d] = "min_"
    assert ob[min_d] == "min_"

    with pytest.raises(TypeError):
        ob = Test.DecimalIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.DecimalIndexerTest()
        ob["wrong"] = "wrong"


def test_string_indexer():
    """Test String indexers."""
    ob = Test.StringIndexerTest()

    assert ob["spam"] is None
    assert ob[u"spam"] is None

    ob["spam"] = "spam"
    assert ob["spam"] == "spam"
    assert ob["spam"] == u"spam"
    assert ob[u"spam"] == "spam"
    assert ob[u"spam"] == u"spam"

    ob[u"eggs"] = u"eggs"
    assert ob["eggs"] == "eggs"
    assert ob["eggs"] == u"eggs"
    assert ob[u"eggs"] == "eggs"
    assert ob[u"eggs"] == u"eggs"

    with pytest.raises(TypeError):
        ob = Test.StringIndexerTest()
        ob[1]

    with pytest.raises(TypeError):
        ob = Test.StringIndexerTest()
        ob[1] = "wrong"


def test_enum_indexer():
    """Test enum indexers."""
    ob = Test.EnumIndexerTest()

    key = Test.ShortEnum.One

    assert ob[key] is None

    ob[key] = "spam"
    assert ob[key] == "spam"

    ob[key] = "eggs"
    assert ob[key] == "eggs"

    with pytest.raises(TypeError):
        ob[1] = "spam"
    with pytest.raises(TypeError):
        ob[1]

    with pytest.raises(TypeError):
        ob = Test.EnumIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.EnumIndexerTest()
        ob["wrong"] = "wrong"


def test_object_indexer():
    """Test ob indexers."""
    ob = Test.ObjectIndexerTest()

    from Python.Test import Spam
    spam = Spam("spam")

    assert ob[spam] is None
    assert ob["spam"] is None
    assert ob[1] is None
    assert ob[None] is None

    ob[spam] = "spam"
    assert ob[spam] == "spam"

    ob["spam"] = "eggs"
    assert ob["spam"] == "eggs"

    ob[1] = "one"
    assert ob[1] == "one"

    ob[1] = "long"
    assert ob[1] == "long"

    class Eggs(object):
        pass

    key = Eggs()
    ob = Test.ObjectIndexerTest()
    ob[key] = "eggs_key"
    assert ob[key] == "eggs_key"


def test_interface_indexer():
    """Test interface indexers."""
    ob = Test.InterfaceIndexerTest()

    from Python.Test import Spam
    spam = Spam("spam")

    assert ob[spam] is None

    ob[spam] = "spam"
    assert ob[spam] == "spam"

    ob[spam] = "eggs"
    assert ob[spam] == "eggs"

    with pytest.raises(TypeError):
        ob = Test.InterfaceIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.InterfaceIndexerTest()
        ob["wrong"] = "wrong"


def test_typed_indexer():
    """Test typed indexers."""
    ob = Test.TypedIndexerTest()

    from Python.Test import Spam
    spam = Spam("spam")

    assert ob[spam] is None

    ob[spam] = "spam"
    assert ob[spam] == "spam"

    ob[spam] = "eggs"
    assert ob[spam] == "eggs"

    with pytest.raises(TypeError):
        ob = Test.TypedIndexerTest()
        ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.TypedIndexerTest()
        ob["wrong"] = "wrong"


def test_multi_arg_indexer():
    """Test indexers that take multiple index arguments."""
    ob = Test.MultiArgIndexerTest()

    ob[0, 1] = "zero one"
    assert ob[0, 1] == "zero one"

    ob[1, 9] = "one nine"
    assert ob[1, 9] == "one nine"

    assert ob[10, 50] is None

    with pytest.raises(TypeError):
        ob = Test.MultiArgIndexerTest()
        _ = ob[0, "one"]

    with pytest.raises(TypeError):
        ob = Test.MultiArgIndexerTest()
        ob[0, "one"] = "wrong"


def test_multi_type_indexer():
    """Test indexers that take multiple indices of different types."""
    ob = Test.MultiTypeIndexerTest()
    spam = Test.Spam("spam")

    ob[0, "one", spam] = "zero one spam"
    assert ob[0, "one", spam] == "zero one spam"

    ob[1, "nine", spam] = "one nine spam"
    assert ob[1, "nine", spam] == "one nine spam"

    with pytest.raises(TypeError):
        ob = Test.MultiTypeIndexerTest()
        _ = ob[0, 1, spam]

    with pytest.raises(TypeError):
        ob = Test.MultiTypeIndexerTest()
        ob[0, 1, spam] = "wrong"


def test_multi_default_key_indexer():
    """Test indexers that take multiple indices with a default
    key arguments."""
    # default argument is 2 in the MultiDefaultKeyIndexerTest object
    ob = Test.MultiDefaultKeyIndexerTest()
    ob[0, 2] = "zero one spam"
    assert ob[0] == "zero one spam"

    ob[1] = "one nine spam"
    assert ob[1, 2] == "one nine spam"


def test_indexer_wrong_key_type():
    """Test calling an indexer using a key of the wrong type."""

    with pytest.raises(TypeError):
        ob = Test.PublicIndexerTest()
        _ = ob["wrong"]

    with pytest.raises(TypeError):
        ob = Test.PublicIndexerTest()
        ob["wrong"] = "spam"


def test_indexer_wrong_value_type():
    """Test calling an indexer using a value of the wrong type."""

    with pytest.raises(TypeError):
        ob = Test.PublicIndexerTest()
        ob[1] = 9993.9


def test_unbound_indexer():
    """Test calling an unbound indexer."""
    ob = Test.PublicIndexerTest()

    Test.PublicIndexerTest.__setitem__(ob, 0, "zero")
    assert ob[0] == "zero"

    Test.PublicIndexerTest.__setitem__(ob, 1, "one")
    assert ob[1] == "one"

    assert ob[10] is None


def test_indexer_abuse():
    """Test indexer abuse."""
    _class = Test.PublicIndexerTest
    ob = Test.PublicIndexerTest()

    with pytest.raises(AttributeError):
        del _class.__getitem__

    with pytest.raises(AttributeError):
        del ob.__getitem__

    with pytest.raises(AttributeError):
        del _class.__setitem__

    with pytest.raises(AttributeError):
        del ob.__setitem__


def test_indexer_accessed_through_interface():
    """Test that indexers can be accessed through interfaces"""
    from System.Collections.Generic import Dictionary, IDictionary
    d = IDictionary[str, str](Dictionary[str, str]())
    d["one"] = "1"
    assert d["one"] == "1"


def test_using_indexer_on_object_without_indexer():
    """Test using subscript syntax on an object an without indexer raises"""
    from System import Object
    o = Object()
    with pytest.raises(TypeError):
        o[0]

    with pytest.raises(TypeError):
        o[0] = 1


def test_inherited_indexer():
    """Test that inherited indexers are accessible"""
    from Python.Test import PublicInheritedIndexerTest
    from Python.Test import ProtectedInheritedIndexerTest
    from Python.Test import PrivateInheritedIndexerTest
    from Python.Test import InternalInheritedIndexerTest

    pub = PublicInheritedIndexerTest()
    pub[0] = "zero"
    assert pub[0] == "zero"

    def assert_no_indexer(obj):
        with pytest.raises(TypeError):
            obj[0]
        with pytest.raises(TypeError):
            obj[0] = "zero"

    assert_no_indexer(PrivateInheritedIndexerTest)
    assert_no_indexer(ProtectedInheritedIndexerTest)
    assert_no_indexer(InternalInheritedIndexerTest)


def test_inherited_indexer_interface():
    """Test that indexers inherited from other interfaces are accessible"""
    from Python.Test import InterfaceInheritedIndexerTest, IInheritedIndexer

    impl = InterfaceInheritedIndexerTest()
    ifc = IInheritedIndexer(impl)
    ifc[0] = "zero"
    assert ifc[0] == "zero"

def test_public_inherited_overloaded_indexer():
    """Test public indexers."""
    ob = Test.PublicInheritedOverloadedIndexer()

    ob[0] = "zero"
    assert ob[0] == "zero"

    ob[1] = "one"
    assert ob[1] == "one"

    assert ob[10] is None

    ob["spam"] = "spam"
    assert ob["spam"] == "spam"

    with pytest.raises(TypeError):
        ob[[]]



#!/usr/bin/env python
# -*- coding: utf-8 -*-
# FIXME: FAIL: testImplicitAssemblyLoad AssertionError: 0 != 1

"""
Run all of the unit tests for this package multiple times in a highly
multi-threaded way to stress the system. This makes it possible to look
for memory leaks and threading issues and provides a good target for a
profiler to accumulate better data.
"""

from __future__ import print_function

import gc
import os
import sys
import threading
import time

import _thread as thread
from .utils import dprint


class StressTest(object):
    def __init__(self):
        self.dirname = os.path.split(__file__)[0]
        sys.path.append(self.dirname)
        gc.set_debug(gc.DEBUG_LEAK)
        import runtests
        self.module = runtests
        self.done = []

    def mark_start(self):
        self._start = time.clock()

    def mark_finish(self):
        self._finish = time.clock()

    def elapsed(self):
        return self._finish - self._start

    def print_gc_report(self):
        for item in gc.get_objects():
            print(item, sys.getrefcount(item))

    def run_thread(self, iterations):
        thread_id = thread.get_ident()
        dprint("thread {0} starting...".format(thread_id))
        time.sleep(0.1)
        for i in range(iterations):
            dprint("thread {0} iter {1} start".format(thread_id, i))
            self.module.main()
            dprint("thread {0} iter {1} end".format(thread_id, i))
        self.done.append(None)
        dprint("thread {0} done".format(thread_id))

    def stress_test(self, iterations=1, threads=1):
        args = (iterations,)
        self.mark_start()
        for _ in range(threads):
            thread = threading.Thread(target=self.run_thread, args=args)
            thread.start()
        while len(self.done) < (iterations * threads):
            dprint(len(self.done))
            time.sleep(0.1)
        self.mark_finish()
        took = self.elapsed()
        self.print_gc_report()


def main():
    test = StressTest()
    test.stress_test(2, 10)


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-

"""Test the import statement."""

import pytest
import sys

def test_relative_missing_import():
    """Test that a relative missing import doesn't crash.
    Some modules use this to check if a package is installed.
    Relative import in the site-packages folder"""
    with pytest.raises(ImportError):
        from . import _missing_import


def test_import_all_on_second_time():
    """Test import all attributes after a normal import without '*'.
    Due to import * only allowed at module level, the test body splited
    to a module file."""
    from . import importtest
    del sys.modules[importtest.__name__]
    
